import collections
import collections.abc as cabc
import copy
import enum
import inspect
import types
import typing as typ
import warnings

import numpy as np
import scipy.sparse.linalg as spsl

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.runtime as pxrt
import pyxu.util as pxu


class Property(enum.Enum):
    """
    Mathematical property.

    See Also
    --------
    :py:class:`~pyxu.abc.Operator`
    """

    CAN_EVAL = enum.auto()
    FUNCTIONAL = enum.auto()
    PROXIMABLE = enum.auto()
    DIFFERENTIABLE = enum.auto()
    DIFFERENTIABLE_FUNCTION = enum.auto()
    LINEAR = enum.auto()
    LINEAR_SQUARE = enum.auto()
    LINEAR_NORMAL = enum.auto()
    LINEAR_IDEMPOTENT = enum.auto()
    LINEAR_SELF_ADJOINT = enum.auto()
    LINEAR_POSITIVE_DEFINITE = enum.auto()
    LINEAR_UNITARY = enum.auto()
    QUADRATIC = enum.auto()

    def arithmetic_methods(self) -> cabc.Set[str]:
        "Instance methods affected by arithmetic operations."
        data = collections.defaultdict(list)
        data[self.CAN_EVAL].extend(
            [
                "apply",
                "__call__",
                "estimate_lipschitz",
                "_expr",
            ]
        )
        data[self.FUNCTIONAL].append("asloss")
        data[self.PROXIMABLE].append("prox")
        data[self.DIFFERENTIABLE].extend(
            [
                "jacobian",
                "estimate_diff_lipschitz",
            ]
        )
        data[self.DIFFERENTIABLE_FUNCTION].append("grad")
        data[self.LINEAR].extend(
            [
                "adjoint",
                "asarray",
                "svdvals",
                "pinv",
                "gram",
                "cogram",
            ]
        )
        data[self.LINEAR_SQUARE].append("trace")
        data[self.QUADRATIC].append("_quad_spec")

        meth = frozenset(data[self])
        return meth


class Operator:
    """
    Abstract Base Class for Pyxu operators.

    Goals:

    * enable operator arithmetic.
    * cast operators to specialized forms.
    * attach :py:class:`~pyxu.abc.Property` tags encoding certain mathematical properties.  Each core sub-class **must**
      have a unique set of properties to be distinguishable from its peers.
    """

    # For `(size-1 ndarray) * OpT` to work, we need to force NumPy's hand and call OpT.__rmul__() in
    # place of ndarray.__mul__() to determine how scaling should be performed.
    # This is achieved by increasing __array_priority__ for all operators.
    __array_priority__ = np.inf

    def __init__(self, shape: pxt.OpShape):
        r"""
        Parameters
        ----------
        shape: OpShape
            (N, M) operator shape.
        """
        shape = pxu.as_canonical_shape(shape)
        assert len(shape) == 2, f"shape: expected {pxt.OpShape}, got {shape}."

        self._shape = shape
        self._name = self.__class__.__name__

    # Public Interface --------------------------------------------------------
    @property
    def shape(self) -> pxt.OpShape:
        r"""
        Return (N, M) operator shape.
        """
        return self._shape

    @property
    def dim(self) -> pxt.Integer:
        r"""
        Return dimension of operator's domain. (M)
        """
        return self.shape[1]

    @property
    def codim(self) -> pxt.Integer:
        r"""
        Return dimension of operator's co-domain. (N)
        """
        return self.shape[0]

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        "Mathematical properties of the operator."
        return frozenset()

    @classmethod
    def has(cls, prop: typ.Union[Property, cabc.Collection[Property]]) -> bool:
        """
        Verify if operator possesses supplied properties.
        """
        if isinstance(prop, Property):
            prop = (prop,)
        return frozenset(prop) <= cls.properties()

    def asop(self, cast_to: pxt.OpC) -> pxt.OpT:
        r"""
        Recast an :py:class:`~pyxu.abc.Operator` (or subclass thereof) to another :py:class:`~pyxu.abc.Operator`.

        Users may call this method if the arithmetic API yields sub-optimal return types.

        This method is a no-op if `cast_to` is a parent class of ``self``.

        Parameters
        ----------
        cast_to: OpC
            Target type for the recast.

        Returns
        -------
        op: OpT
            Operator with the new interface.

            Fails when cast is forbidden.
            (Ex: :py:class:`~pyxu.abc.Map` -> :py:class:`~pyxu.abc.Func` if codim > 1)

        Notes
        -----
        * The interface of `cast_to` is provided via encapsulation + forwarding.
        * If ``self`` does not implement all methods from `cast_to`, then unimplemented methods will raise
          :py:class:`NotImplementedError` when called.
        """
        if cast_to not in _core_operators():
            raise ValueError(f"cast_to: expected a core base-class, got {cast_to}.")

        p_core = frozenset(self.properties())
        p_shell = frozenset(cast_to.properties())
        if p_shell <= p_core:
            # Trying to cast `self` to it's own class or a parent class.
            # Inheritance rules mean the target object already satisfies the intended interface.
            return self
        else:
            # (p_shell > p_core) -> specializing to a sub-class of ``self``
            # OR
            # len(p_shell ^ p_core) > 0 -> specializing to another branch of the class hierarchy.
            op = cast_to(shape=self.shape)
            op._core = self  # for debugging

            # Forward shared arithmetic fields from core to shell.
            for p in p_shell & p_core:
                for m in p.arithmetic_methods():
                    m_core = getattr(self, m)
                    setattr(op, m, m_core)

            # [diff_]lipschitz are not arithmetic methods, hence are not propagated.
            # We propagate them manually to avoid un-warranted re-evaluations.
            # Important: we write to _[diff_]lipschitz to not overwrite estimate_[diff_]lipschitz() methods.
            if cast_to.has(Property.CAN_EVAL) and self.has(Property.CAN_EVAL):
                op._lipschitz = self.lipschitz
            if cast_to.has(Property.DIFFERENTIABLE) and self.has(Property.DIFFERENTIABLE):
                op._diff_lipschitz = self.diff_lipschitz

            return op

    # Operator Arithmetic -----------------------------------------------------
    def __add__(self, other: pxt.OpT) -> pxt.OpT:
        """
        Add two operators.

        Parameters
        ----------
        self: OpT
            (A, B) Left operand.
        other: OpT
            (C, D) Right operand.

        Returns
        -------
        op: OpT
            Composite operator ``self + other``

        Notes
        -----
        Operand shapes must be `consistent`, i.e.:

            * have `identical shape`, or
            * be `range-broadcastable`, i.e. functional + map works.

        See Also
        --------
        :py:class:`~pyxu.abc.arithmetic.AddRule`
        """
        import pyxu.abc.arithmetic as arithmetic

        if isinstance(other, Operator):
            return arithmetic.AddRule(lhs=self, rhs=other).op()
        else:
            return NotImplemented

    def __sub__(self, other: pxt.OpT) -> pxt.OpT:
        """
        Subtract two operators.

        Parameters
        ----------
        self: OpT
            (A, B) Left operand.
        other: OpT
            (C, D) Right operand.

        Returns
        -------
        op: OpT
            Composite operator ``self - other``
        """
        import pyxu.abc.arithmetic as arithmetic

        if isinstance(other, Operator):
            return arithmetic.AddRule(lhs=self, rhs=-other).op()
        else:
            return NotImplemented

    def __neg__(self) -> pxt.OpT:
        """
        Negate an operator.

        Returns
        -------
        op: OpT
            Composite operator ``-1 * self``.
        """
        import pyxu.abc.arithmetic as arithmetic

        return arithmetic.ScaleRule(op=self, cst=-1).op()

    def __mul__(self, other: typ.Union[pxt.Real, pxt.OpT]) -> pxt.OpT:
        """
        Compose two operators, or scale an operator by a constant.

        Parameters
        ----------
        self: OpT
            (A, B) Left operand.
        other: Real, OpT
            (1,) scalar, or
            (C, D) Right operand.

        Returns
        -------
        op: OpT
            (A, B) scaled operator, or
            (A, D) composed operator ``self * other``.

        Notes
        -----
        If called with two operators, their shapes must be `consistent`, i.e. B == C.

        See Also
        --------
        :py:class:`~pyxu.abc.arithmetic.ScaleRule`,
        :py:class:`~pyxu.abc.arithmetic.ChainRule`
        """
        import pyxu.abc.arithmetic as arithmetic

        if isinstance(other, Operator):
            return arithmetic.ChainRule(lhs=self, rhs=other).op()
        elif _is_real(other):
            return arithmetic.ScaleRule(op=self, cst=float(other)).op()
        else:
            return NotImplemented

    def __rmul__(self, other: pxt.Real) -> pxt.OpT:
        import pyxu.abc.arithmetic as arithmetic

        if _is_real(other):
            return arithmetic.ScaleRule(op=self, cst=float(other)).op()
        else:
            return NotImplemented

    def __truediv__(self, other: pxt.Real) -> pxt.OpT:
        import pyxu.abc.arithmetic as arithmetic

        if _is_real(other):
            return arithmetic.ScaleRule(op=self, cst=float(1 / other)).op()
        else:
            return NotImplemented

    def __pow__(self, k: pxt.Integer) -> pxt.OpT:
        """
        Exponentiate an operator, i.e. compose it with itself.

        Parameters
        ----------
        k: Integer
            Number of times the operator is composed with itself.

        Returns
        -------
        op: OpT
            Exponentiated operator.

        Notes
        -----
        Exponentiation is only allowed for endomorphisms, i.e. square operators.

        See Also
        --------
        :py:class:`~pyxu.abc.arithmetic.PowerRule`
        """
        import pyxu.abc.arithmetic as arithmetic

        if isinstance(k, pxt.Integer) and (k >= 0):
            return arithmetic.PowerRule(op=self, k=k).op()
        else:
            return NotImplemented

    def __matmul__(self, other) -> pxt.OpT:
        # (op @ NDArray) unsupported
        return NotImplemented

    def __rmatmul__(self, other) -> pxt.OpT:
        # (NDArray @ op) unsupported
        return NotImplemented

    def argscale(self, scalar: pxt.Real) -> pxt.OpT:
        """
        Scale operator's domain.

        Parameters
        ----------
        scalar: Real

        Returns
        -------
        op: OpT
            (N, M) domain-scaled operator.

        See Also
        --------
        :py:class:`~pyxu.abc.arithmetic.ArgScaleRule`
        """
        import pyxu.abc.arithmetic as arithmetic

        assert _is_real(scalar)
        return arithmetic.ArgScaleRule(op=self, cst=float(scalar)).op()

    def argshift(self, shift: typ.Union[pxt.Real, pxt.NDArray]) -> pxt.OpT:
        """
        Shift operator's domain.

        Parameters
        ----------
        shift: Real, NDArray
            (M,) shift value

        Returns
        -------
        op: OpT
            (N, M) domain-shifted operator.

        See Also
        --------
        :py:class:`~pyxu.abc.arithmetic.ArgShiftRule`
        """
        import pyxu.abc.arithmetic as arithmetic

        if _is_real(shift):
            shift = float(shift)
        return arithmetic.ArgShiftRule(op=self, cst=shift).op()

    # Internal Helpers --------------------------------------------------------
    @staticmethod
    def _infer_operator_type(prop: cabc.Collection[Property]) -> pxt.OpC:
        prop = frozenset(prop)
        for op in _core_operators():
            if op.properties() == prop:
                return op
        else:
            raise ValueError(f"No operator found with properties {prop}.")

    def squeeze(self) -> pxt.OpT:
        r"""
        Cast an :py:class:`~pyxu.abc.Operator` to the right core operator sub-type given codomain dimension.
        """
        p = set(self.properties())
        if self.codim == 1:
            p.add(Property.FUNCTIONAL)
            if Property.DIFFERENTIABLE in self.properties():
                p.add(Property.DIFFERENTIABLE_FUNCTION)
            if Property.LINEAR in self.properties():
                for p_ in Property:
                    if p_.name.startswith("LINEAR_"):
                        p.discard(p_)
                p.add(Property.PROXIMABLE)
        elif self.codim == self.dim:
            if Property.LINEAR in self.properties():
                p.add(Property.LINEAR_SQUARE)
        klass = self._infer_operator_type(p)
        return self.asop(klass)

    def __repr__(self) -> str:
        klass = self._name
        return f"{klass}{self.shape}"

    def _expr(self) -> tuple:
        r"""
        Show the expression-representation of the operator.

        If overridden, must return a tuple of the form

            (head, \*tail),

        where `head` is the operator (ex: +/\*), and `tail` denotes all the expression's terms.  If an operator cannot
        be expanded further, then this method should return (self,).
        """
        return (self,)

    def expr(self, level: int = 0, strip: bool = True) -> str:
        """
        Pretty-Print the expression representation of the operator.

        Useful for debugging arithmetic-induced expressions.

        Example
        -------

        .. code-block:: python3

           import numpy as np
           import pyxu.abc as pxa

           N = 5
           op1 = pxa.LinFunc.from_array(np.arange(N))
           op2 = pxa.LinOp.from_array(np.ones((N, N)))
           op = ((2 * op1) + (op2 ** 3)).argshift(np.full(N, 4))

           print(op.expr())
           # [argshift, ==> DiffMap(5, 5)
           # .[add, ==> SquareOp(5, 5)
           # ..[scale, ==> LinFunc(1, 5)
           # ...LinFunc(1, 5),
           # ...2.0],
           # ..[exp, ==> SquareOp(5, 5)
           # ...LinOp(5, 5),
           # ...3]],
           # .(5,)]
        """
        fmt = lambda obj, lvl: ("." * lvl) + str(obj)
        lines = []

        head, *tail = self._expr()
        if len(tail) == 0:
            head = f"{repr(head)},"
        else:
            head = f"[{head}, ==> {repr(self)}"
        lines.append(fmt(head, level))

        for t in tail:
            if isinstance(t, Operator):
                lines += t.expr(level=level + 1, strip=False).split("\n")
            else:
                t = f"{t},"
                lines.append(fmt(t, level + 1))
        if len(tail) > 0:
            # Drop comma for last tail item, then close the sub-expression.
            lines[-1] = lines[-1][:-1]
            lines[-1] += "],"

        out = "\n".join(lines)
        if strip:
            out = out.strip(",")  # drop comma at top-level tail.
        return out


class Map(Operator):
    r"""
    Base class for real-valued maps :math:`\mathbf{f}: \mathbb{R}^{M} \to \mathbb{R}^{N}`.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply`.

    If :math:`\mathbf{f}` is Lipschitz-continuous with known Lipschitz constant :math:`L`, the latter should be stored
    in the :py:attr:`~pyxu.abc.Map.lipschitz` property.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.CAN_EVAL)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        self.lipschitz = np.inf

    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        """
        Evaluate operator at specified point(s).

        Parameters
        ----------
        arr: NDArray
            (..., M) input points.

        Returns
        -------
        out: NDArray
            (..., N) output points.
        """
        raise NotImplementedError

    def __call__(self, arr: pxt.NDArray) -> pxt.NDArray:
        """
        Alias for :py:meth:`~pyxu.abc.Map.apply`.
        """
        return self.apply(arr)

    @property
    def lipschitz(self) -> pxt.Real:
        r"""
        Return the last computed Lipschitz constant of :math:`\mathbf{f}`.

        Notes
        -----
        * If a Lipschitz constant is known apriori, it can be stored in the instance as follows:

          .. code-block:: python3

             class TestOp(Map):
                 def __init__(self, shape):
                     super().__init__(shape=shape)
                     self.lipschitz = 2

             op = TestOp(shape=(2, 3))
             op.lipschitz  # => 2


          Alternatively the Lipschitz constant can be set manually after initialization:

          .. code-block:: python3

             class TestOp(Map):
                 def __init__(self, shape):
                     super().__init__(shape=shape)

             op = TestOp(shape=(2,3))
             op.lipschitz  # => inf, since unknown apriori

             op.lipschitz = 2  # post-init specification
             op.lipschitz  # => 2

        * :py:meth:`~pyxu.abc.Map.lipschitz` **never** computes anything:
          call :py:meth:`~pyxu.abc.Map.estimate_lipschitz` manually to *compute* a new Lipschitz estimate:

          .. code-block:: python3

             op.lipschitz = op.estimate_lipschitz()
        """
        if not hasattr(self, "_lipschitz"):
            self._lipschitz = self.estimate_lipschitz()

        return pxrt.coerce(self._lipschitz)

    @lipschitz.setter
    def lipschitz(self, L: pxt.Real):
        assert L >= 0
        self._lipschitz = float(L)

        # If no algorithm available to auto-determine estimate_lipschitz(), then enforce user's choice.
        if not self.has(Property.LINEAR):

            def op_estimate_lipschitz(_, **kwargs) -> pxt.Real:
                return _._lipschitz

            self.estimate_lipschitz = types.MethodType(op_estimate_lipschitz, self)

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        r"""
        Compute a Lipschitz constant of the operator.

        Parameters
        ----------
        kwargs: ~collections.abc.Mapping
            Class-specific kwargs to configure Lipschitz estimation.

        Notes
        -----
        * This method should always be callable without specifying any kwargs.

        * A constant :math:`L_{\mathbf{f}} > 0` is said to be a *Lipschitz constant* for a map :math:`\mathbf{f}:
          \mathbb{R}^{M} \to \mathbb{R}^{N}` if:

          .. math::

             \|\mathbf{f}(\mathbf{x}) - \mathbf{f}(\mathbf{y})\|_{\mathbb{R}^{N}}
             \leq
             L_{\mathbf{f}} \|\mathbf{x} - \mathbf{y}\|_{\mathbb{R}^{M}},
             \qquad
             \forall \mathbf{x}, \mathbf{y}\in \mathbb{R}^{M},

          where :math:`\|\cdot\|_{\mathbb{R}^{M}}` and :math:`\|\cdot\|_{\mathbb{R}^{N}}` are the canonical norms on
          their respective spaces.

          The smallest Lipschitz constant of a map is called the *optimal Lipschitz constant*.
        """
        raise NotImplementedError


class Func(Map):
    r"""
    Base class for real-valued functionals :math:`f: \mathbb{R}^{M} \to \mathbb{R}\cup\{+\infty\}`.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply`.

    If :math:`f` is Lipschitz-continuous with known Lipschitz constant :math:`L`, the latter should be stored in the
    :py:attr:`~pyxu.abc.Map.lipschitz` property.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.FUNCTIONAL)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        assert self.codim == 1, f"shape: expected (1, n), got {shape}."

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        """
        Transform a functional into a loss functional.

        The meaning of :py:meth:`~pyxu.abc.Func.asloss` is operator-dependent: always examine the output of
        :py:meth:`~pyxu.abc.Operator.expr` to determine the loss-notion involved.

        Parameters
        ----------
        data: NDArray
            (M,) input.

        Returns
        -------
        op: OpT
            (1, M) loss function.
            If `data` is unspecified, returns ``self``.
        """
        raise NotImplementedError


class DiffMap(Map):
    r"""
    Base class for real-valued differentiable maps :math:`\mathbf{f}: \mathbb{R}^{M} \to \mathbb{R}^{N}`.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply` and :py:meth:`~pyxu.abc.DiffMap.jacobian`.

    If :math:`\mathbf{f}` is Lipschitz-continuous with known Lipschitz constant :math:`L`, the latter should be stored
    in the :py:attr:`~pyxu.abc.Map.lipschitz` property.

    If :math:`\mathbf{J}_{\mathbf{f}}` is Lipschitz-continuous with known Lipschitz constant :math:`\partial L`, the
    latter should be stored in the :py:attr:`~pyxu.abc.DiffMap.diff_lipschitz` property.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.DIFFERENTIABLE)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        self.diff_lipschitz = np.inf

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        r"""
        Evaluate the Jacobian of :math:`\mathbf{f}` at the specified point.

        Parameters
        ----------
        arr: NDArray
            (M,) evaluation point.

        Returns
        -------
        op: OpT
            (N, M) Jacobian operator at point `arr`.

        Notes
        -----
        Let :math:`\mathbf{f}=[f_{1}, \ldots, f_{N}]: \mathbb{R}^{M} \to \mathbb{R}^{N}` be a differentiable
        multi-dimensional map.  The *Jacobian* (or *differential*) of :math:`\mathbf{f}` at :math:`\mathbf{z} \in
        \mathbb{R}^{M}` is defined as the best linear approximator of :math:`\mathbf{f}` near :math:`\mathbf{z}`, in the
        following sense:

        .. math::

           \mathbf{f}(\mathbf{x}) - \mathbf{f}(\mathbf{z})
           =
           \mathbf{J}_{\mathbf{f}}(\mathbf{z}) (\mathbf{x} - \mathbf{z})
           +
           o(\| \mathbf{x} - \mathbf{z} \|)
           \quad
           \text{as} \quad \mathbf{x} \to \mathbf{z}.

        The Jacobian admits the following matrix representation:

        .. math::

           [\mathbf{J}_{\mathbf{f}}(\mathbf{x})]_{ij}
           :=
           \frac{\partial f_{i}}{\partial x_{j}}(\mathbf{x}),
           \qquad
           \forall (i,j) \in \{1,\ldots,N\} \times \{1,\ldots,M\}.
        """
        raise NotImplementedError

    @property
    def diff_lipschitz(self) -> pxt.Real:
        r"""
        Return the last computed Lipschitz constant of :math:`\mathbf{J}_{\mathbf{f}}`.

        Notes
        -----
        * If a diff-Lipschitz constant is known apriori, it can be stored in the instance as follows:

          .. code-block:: python3

             class TestOp(DiffMap):
                 def __init__(self, shape):
                     super().__init__(shape=shape)
                     self.diff_lipschitz = 2

             op = TestOp(shape=(2, 3))
             op.diff_lipschitz  # => 2


          Alternatively the diff-Lipschitz constant can be set manually after initialization:

          .. code-block:: python3

             class TestOp(DiffMap):
                 def __init__(self, shape):
                     super().__init__(shape=shape)

             op = TestOp(shape=(2,3))
             op.diff_lipschitz  # => inf, since unknown apriori

             op.diff_lipschitz = 2  # post-init specification
             op.diff_lipschitz  # => 2

        * :py:meth:`~pyxu.abc.DiffMap.diff_lipschitz` **never** computes anything:
          call :py:meth:`~pyxu.abc.DiffMap.estimate_diff_lipschitz` manually to *compute* a new diff-Lipschitz estimate:

          .. code-block:: python3

             op.diff_lipschitz = op.estimate_diff_lipschitz()
        """
        if not hasattr(self, "_diff_lipschitz"):
            self._diff_lipschitz = self.estimate_diff_lipschitz()

        return pxrt.coerce(self._diff_lipschitz)

    @diff_lipschitz.setter
    def diff_lipschitz(self, dL: pxt.Real):
        assert dL >= 0
        self._diff_lipschitz = float(dL)

        # If no algorithm available to auto-determine estimate_diff_lipschitz(), then enforce user's choice.
        if not self.has(Property.QUADRATIC):

            def op_estimate_diff_lipschitz(_, **kwargs) -> pxt.Real:
                return _._diff_lipschitz

            self.estimate_diff_lipschitz = types.MethodType(op_estimate_diff_lipschitz, self)

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        r"""
        Compute a Lipschitz constant of :py:meth:`~pyxu.abc.DiffMap.jacobian`.

        Parameters
        ----------
        kwargs: ~collections.abc.Mapping
            Class-specific kwargs to configure diff-Lipschitz estimation.

        Notes
        -----
        * This method should always be callable without specifying any kwargs.

        * A Lipschitz constant :math:`L_{\mathbf{J}_{\mathbf{f}}} > 0` of the Jacobian map
          :math:`\mathbf{J}_{\mathbf{f}}: \mathbb{R}^{M} \to \mathbb{R}^{N \times M}` is such that:

          .. math::

             \|\mathbf{J}_{\mathbf{f}}(\mathbf{x}) - \mathbf{J}_{\mathbf{f}}(\mathbf{y})\|_{\mathbb{R}^{N \times M}}
             \leq
             L_{\mathbf{J}_{\mathbf{f}}} \|\mathbf{x} - \mathbf{y}\|_{\mathbb{R}^{M}},
             \qquad
             \forall \mathbf{x}, \mathbf{y} \in \mathbb{R}^{M},

          where :math:`\|\cdot\|_{\mathbb{R}^{N \times M}}` and :math:`\|\cdot\|_{\mathbb{R}^{M}}` are the canonical
          norms on their respective spaces.

          The smallest Lipschitz constant of the Jacobian is called the *optimal diff-Lipschitz constant*.
        """
        raise NotImplementedError


class ProxFunc(Func):
    r"""
    Base class for real-valued proximable functionals :math:`f: \mathbb{R}^{M} \to \mathbb{R} \cup \{+\infty\}`.

    A functional :math:`f: \mathbb{R}^{M} \to \mathbb{R} \cup \{+\infty\}` is said *proximable* if its **proximity
    operator** (see :py:meth:`~pyxu.abc.ProxFunc.prox` for a definition) admits a *simple closed-form expression* **or**
    can be evaluated *efficiently* and with *high accuracy*.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply` and :py:meth:`~pyxu.abc.ProxFunc.prox`.

    If :math:`f` is Lipschitz-continuous with known Lipschitz constant :math:`L`, the latter should be stored in the
    :py:attr:`~pyxu.abc.Map.lipschitz` property.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.PROXIMABLE)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)

    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        r"""
        Evaluate proximity operator of :math:`\tau f` at specified point(s).

        Parameters
        ----------
        arr: NDArray
            (..., M) input points.
        tau: Real
            Positive scale factor.

        Returns
        -------
        out: NDArray
            (..., M) proximal evaluations.

        Notes
        -----
        For :math:`\tau >0`, the *proximity operator* of a scaled functional :math:`f: \mathbb{R}^{M} \to \mathbb{R}` is
        defined as:

        .. math::

           \mathbf{\text{prox}}_{\tau f}(\mathbf{z})
           :=
           \arg\min_{\mathbf{x}\in\mathbb{R}^{M}} f(x)+\frac{1}{2\tau} \|\mathbf{x}-\mathbf{z}\|_{2}^{2},
           \quad
           \forall \mathbf{z} \in \mathbb{R}^{M}.
        """
        raise NotImplementedError

    @pxrt.enforce_precision(i=("arr", "sigma"))
    def fenchel_prox(self, arr: pxt.NDArray, sigma: pxt.Real) -> pxt.NDArray:
        r"""
        Evaluate proximity operator of :math:`\sigma f^{\ast}`, the scaled Fenchel conjugate of :math:`f`, at specified
        point(s).

        Parameters
        ----------
        arr: NDArray
            (..., M) input points.
        sigma: Real
            Positive scale factor.

        Returns
        -------
        out: NDArray
            (..., M) proximal evaluations.

        Notes
        -----
        For :math:`\sigma > 0`, the *Fenchel conjugate* is defined as:

        .. math::

           f^{\ast}(\mathbf{z})
           :=
           \max_{\mathbf{x}\in\mathbb{R}^{M}} \langle \mathbf{x},\mathbf{z} \rangle - f(\mathbf{x}).

        From *Moreau's identity*, its proximal operator is given by:

        .. math::

           \mathbf{\text{prox}}_{\sigma f^{\ast}}(\mathbf{z})
           =
           \mathbf{z} - \sigma \mathbf{\text{prox}}_{f/\sigma}(\mathbf{z}/\sigma).
        """
        out = self.prox(arr=arr / sigma, tau=1 / sigma)
        out = pxu.copy_if_unsafe(out)
        out *= -sigma
        out += arr
        return out

    def moreau_envelope(self, mu: pxt.Real) -> pxt.OpT:
        r"""
        Approximate proximable functional :math:`f` by its *Moreau envelope* :math:`f^{\mu}`.

        Parameters
        ----------
        mu: Real
            Positive regularization parameter.

        Returns
        -------
        op: OpT
            Differential Moreau envelope.

        Notes
        -----
        Consider a convex non-smooth proximable functional :math:`f: \mathbb{R}^{M} \to \mathbb{R} \cup \{+\infty\}` and
        a regularization parameter :math:`\mu > 0`.  The :math:`\mu`-*Moreau envelope* (or *Moreau-Yoshida envelope*) of
        :math:`f` is given by

        .. math::

           f^{\mu}(\mathbf{x})
           =
           \min_{\mathbf{z} \in \mathbb{R}^{M}}
           f(\mathbf{z})
           \quad + \quad
           \frac{1}{2\mu} \|\mathbf{x} - \mathbf{z}\|^{2}.

        The parameter :math:`\mu` controls the trade-off between the regularity properties of :math:`f^{\mu}` and the
        approximation error incurred by the Moreau-Yoshida regularization.

        The Moreau envelope inherits the convexity of :math:`f` and is gradient-Lipschitz (with Lipschitz constant
        :math:`\mu^{-1}`), even if :math:`f` is non-smooth.  Its gradient is moreover given by:

        .. math::

           \nabla f^{\mu}(\mathbf{x})
           =
           \mu^{-1} \left(\mathbf{x} - \text{prox}_{\mu f}(\mathbf{x})\right).

        In addition, :math:`f^{\mu}` envelopes :math:`f` from below: :math:`f^{\mu}(\mathbf{x}) \leq f(\mathbf{x})`.
        This envelope becomes tighter as :math:`\mu \to 0`:

        .. math::

           \lim_{\mu \to 0} f^{\mu}(\mathbf{x}) = f(\mathbf{x}).

        Finally, it can be shown that the minimizers of :math:`f` and :math:`f^{\mu}` coincide, and that the Fenchel
        conjugate of :math:`f^{\mu}` is strongly-convex.

        Example
        -------
        Construct and plot the Moreau envelope of the :math:`\ell_{1}`-norm:

        .. plot::

           import numpy as np
           import matplotlib.pyplot as plt
           from pyxu.abc import ProxFunc

           class L1Norm(ProxFunc):
               def __init__(self, dim: int):
                   super().__init__(shape=(1, dim))
               def apply(self, arr):
                   return np.linalg.norm(arr, axis=-1, keepdims=True, ord=1)
               def prox(self, arr, tau):
                   return np.clip(np.abs(arr)-tau, a_min=0, a_max=None) * np.sign(arr)

           N = 512
           l1_norm = L1Norm(dim=N)
           mus = [0.1, 0.5, 1]
           smooth_l1_norms = [l1_norm.moreau_envelope(mu) for mu in mus]

           x = np.linspace(-1, 1, N)[:, None]
           labels=['mu=0']
           labels.extend([f'mu={mu}' for mu in mus])
           plt.figure()
           plt.plot(x, l1_norm(x))
           for f in smooth_l1_norms:
               plt.plot(x, f(x))
           plt.legend(labels)
           plt.title('Moreau Envelope')

           labels=[f'mu={mu}' for mu in mus]
           plt.figure()
           for f in smooth_l1_norms:
               plt.plot(x, f.grad(x))
           plt.legend(labels)
           plt.title('Derivative of Moreau Envelope')
        """
        from pyxu.operator.interop import from_source

        assert mu > 0, f"mu: expected positive, got {mu}"

        @pxrt.enforce_precision(i="arr")
        def op_apply(_, arr):
            from pyxu.math import norm

            x = self.prox(arr, tau=_._mu)
            out = pxu.copy_if_unsafe(self.apply(x))
            out += (0.5 / _._mu) * norm(arr - x, axis=-1, keepdims=True) ** 2
            return out

        @pxrt.enforce_precision(i="arr")
        def op_grad(_, arr):
            x = arr.copy()
            x -= self.prox(arr, tau=_._mu)
            x /= _._mu
            return x

        op = from_source(
            cls=DiffFunc,
            shape=self.shape,
            embed=dict(
                _name="moreau_envelope",
                _mu=mu,
                _diff_lipschitz=float(1 / mu),
            ),
            apply=op_apply,
            grad=op_grad,
            _expr=lambda _: ("moreau_envelope", _, _._mu),
        )
        return op


class DiffFunc(DiffMap, Func):
    r"""
    Base class for real-valued differentiable functionals :math:`f: \mathbb{R}^{M} \to \mathbb{R}`.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply` and :py:meth:`~pyxu.abc.DiffFunc.grad`.

    If :math:`f` and/or its derivative :math:`f'` are Lipschitz-continuous with known Lipschitz constants :math:`L` and
    :math:`\partial L`, the latter should be stored in the :py:attr:`~pyxu.abc.Map.lipschitz` and
    :py:attr:`~pyxu.abc.DiffMap.diff_lipschitz` properties.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set()
        for klass in cls.__bases__:
            p |= klass.properties()
        p.add(Property.DIFFERENTIABLE_FUNCTION)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        DiffMap.__init__(self, shape)
        Func.__init__(self, shape)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        return LinFunc.from_array(self.grad(arr))

    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        r"""
        Evaluate operator gradient at specified point(s).

        Parameters
        ----------
        arr: NDArray
            (..., M) input points.

        Returns
        -------
        out: NDArray
            (..., M) gradients.

        Notes
        -----
        The gradient of a functional :math:`f: \mathbb{R}^{M} \to \mathbb{R}` is given, for every :math:`\mathbf{x} \in
        \mathbb{R}^{M}`, by

        .. math::

           \nabla f(\mathbf{x})
           :=
           \left[\begin{array}{c}
           \frac{\partial f}{\partial x_{1}}(\mathbf{x}) \\
           \vdots \\
           \frac{\partial f}{\partial x_{M}}(\mathbf{x})
           \end{array}\right].
        """
        raise NotImplementedError


class ProxDiffFunc(ProxFunc, DiffFunc):
    r"""
    Base class for real-valued differentiable *and* proximable functionals :math:`f:\mathbb{R}^{M} \to \mathbb{R}`.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply`, :py:meth:`~pyxu.abc.DiffFunc.grad`, and
    :py:meth:`~pyxu.abc.ProxFunc.prox`.

    If :math:`f` and/or its derivative :math:`f'` are Lipschitz-continuous with known Lipschitz constants :math:`L` and
    :math:`\partial L`, the latter should be stored in the :py:attr:`~pyxu.abc.Map.lipschitz` and
    :py:attr:`~pyxu.abc.DiffMap.diff_lipschitz` properties.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set()
        for klass in cls.__bases__:
            p |= klass.properties()
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        ProxFunc.__init__(self, shape)
        DiffFunc.__init__(self, shape)


class QuadraticFunc(ProxDiffFunc):
    # This is a special abstract base class with more __init__ parameters than `shape`.
    r"""
    Base class for quadratic functionals :math:`f: \mathbb{R}^{M} \to \mathbb{R} \cup \{+\infty\}`.

    The quadratic functional is defined as:

    .. math::

       f(\mathbf{x})
       =
       \frac{1}{2} \langle\mathbf{x}, \mathbf{Q}\mathbf{x}\rangle
       +
       \mathbf{c}^T\mathbf{x}
       +
       t,
       \qquad \forall \mathbf{x} \in \mathbb{R}^{M},

    where :math:`Q` is a positive-definite operator :math:`\mathbf{Q}:\mathbb{R}^{M} \to \mathbb{R}^{M}`,
    :math:`\mathbf{c} \in \mathbb{R}^{M}`, and :math:`t > 0`.

    Its gradient is given by:

    .. math::

       \nabla f(\mathbf{x}) = \mathbf{Q}\mathbf{x} + \mathbf{c}.

    Its proximity operator by:

    .. math::

       \text{prox}_{\tau f}(x)
       =
       \left(
           \mathbf{Q} + \tau^{-1} \mathbf{Id}
       \right)^{-1}
       \left(
           \tau^{-1}\mathbf{x} - \mathbf{c}
       \right).

    In practice the proximity operator is evaluated via :py:class:`~pyxu.opt.solver.CG`.

    The Lipschitz constant :math:`L` of a quadratic on an unbounded domain is unbounded.  The Lipschitz constant
    :math:`\partial L` of :math:`\nabla f` is given by the spectral norm of :math:`\mathbf{Q}`.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.QUADRATIC)
        return frozenset(p)

    def __init__(
        self,
        shape: pxt.OpShape,
        # required in place of `dim` to have uniform interface with Operator hierarchy.
        Q: "PosDefOp" = None,
        c: "LinFunc" = None,
        t: pxt.Real = 0,
    ):
        r"""
        Parameters
        ----------
        Q: ~pyxu.abc.PosDefOp
            (N, N) positive-definite operator. (Default: Identity)
        c: ~pyxu.abc.LinFunc
            (1, N) linear functional. (Default: NullFunc)
        t: Real
            offset. (Default: 0)
        """
        from pyxu.operator import IdentityOp, NullFunc

        super().__init__(shape=shape)

        # Do NOT access (_Q, _c, _t) directly through `self`:
        # their values may not reflect the true (Q, c, t) parameterization.
        # (Reason: arithmetic propagation.)
        # Always access (Q, c, t) by querying the arithmetic method `_quad_spec()`.
        self._Q = IdentityOp(dim=self.dim) if (Q is None) else Q
        self._c = NullFunc(dim=self.dim) if (c is None) else c
        self._t = t

        # ensure dimensions are consistent if None-initialized
        assert self._Q.shape == (self.dim, self.dim)
        assert self._c.shape == self.shape

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        Q, c, t = self._quad_spec()
        out = (arr * Q.apply(arr)).sum(axis=-1, keepdims=True)
        out /= 2
        out += c.apply(arr)
        out += t
        return out

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        Q, c, _ = self._quad_spec()
        out = pxu.copy_if_unsafe(Q.apply(arr))
        out += c.grad(arr)
        return out

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        from pyxu.operator import HomothetyOp
        from pyxu.opt.solver import CG
        from pyxu.opt.stop import MaxIter

        Q, c, _ = self._quad_spec()
        A = Q + HomothetyOp(cst=1 / tau, dim=Q.dim)
        b = arr.copy()
        b /= tau
        b -= c.grad(arr)

        slvr = CG(A=A, show_progress=False)

        sentinel = MaxIter(n=2 * A.dim)
        stop_crit = slvr.default_stop_crit() | sentinel

        slvr.fit(b=b, stop_crit=stop_crit)
        return slvr.solution()

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        from pyxu.operator import shift_loss

        op = shift_loss(op=self, data=data)
        return op

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        Q, *_ = self._quad_spec()
        dL = Q.estimate_lipschitz(**kwargs)
        return dL

    def _quad_spec(self):
        """
        Canonical quadratic parameterization.

        Useful for some internal methods, and overloaded during operator arithmetic.
        """
        return (self._Q, self._c, self._t)


class LinOp(DiffMap):
    r"""
    Base class for real-valued linear operators :math:`\mathbf{A}: \mathbb{R}^{M} \to \mathbb{R}^{N}`.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply` and :py:meth:`~pyxu.abc.LinOp.adjoint`.

    If known, the Lipschitz constant :math:`L` should be stored in the :py:attr:`~pyxu.abc.Map.lipschitz` property.

    The Jacobian of a linear map :math:`\mathbf{A}` is constant.
    """

    # Internal Helpers ------------------------------------
    @staticmethod
    def _warn_vals_sparse_gpu():
        msg = "\n".join(
            [
                "Potential Error:",
                "Sparse GPU-evaluation of svdvals() is known to produce incorrect results. (CuPy-specific + Matrix-Dependant.)",
                "It is advised to cross-check results with CPU-computed results.",
            ]
        )
        warnings.warn(msg, pxw.BackendWarning)

    # -----------------------------------------------------

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.LINEAR)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        self.diff_lipschitz = 0

    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        r"""
        Evaluate operator adjoint at specified point(s).

        Parameters
        ----------
        arr: NDArray
            (..., N) input points.

        Returns
        -------
        out: NDArray
            (..., M) adjoint evaluations.

        Notes
        -----
        The *adjoint* :math:`\mathbf{A}^{\ast}: \mathbb{R}^{N} \to \mathbb{R}^{M}` of :math:`\mathbf{A}: \mathbb{R}^{M}
        \to \mathbb{R}^{N}` is defined as:

        .. math::

           \langle \mathbf{x}, \mathbf{A}^{\ast}\mathbf{y}\rangle_{\mathbb{R}^{M}}
           :=
           \langle \mathbf{A}\mathbf{x}, \mathbf{y}\rangle_{\mathbb{R}^{N}},
           \qquad
           \forall (\mathbf{x},\mathbf{y})\in \mathbb{R}^{M} \times \mathbb{R}^{N}.
        """
        raise NotImplementedError

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        return self

    @property
    def T(self) -> pxt.OpT:
        r"""
        Return the (M, N) adjoint of :math:`\mathbf{A}`.
        """
        import pyxu.abc.arithmetic as arithmetic

        return arithmetic.TransposeRule(op=self).op()

    def to_sciop(
        self,
        dtype: pxt.DType = None,
        gpu: bool = False,
    ) -> spsl.LinearOperator:
        r"""
        Cast a :py:class:`~pyxu.abc.LinOp` to a CPU/GPU :py:class:`~scipy.sparse.linalg.LinearOperator`, compatible with
        the matrix-free linear algebra routines of :py:mod:`scipy.sparse.linalg`.

        Parameters
        ----------
        dtype: DType
            Working precision of the linear operator.
        gpu: bool
            Operate on CuPy inputs (True) vs. NumPy inputs (False).

        Returns
        -------
        op: ~scipy.sparse.linalg.LinearOperator
            Linear operator object compliant with SciPy's interface.
        """

        def matmat(arr):
            with pxrt.EnforcePrecision(False):
                return self.apply(arr.T).T

        def rmatmat(arr):
            with pxrt.EnforcePrecision(False):
                return self.adjoint(arr.T).T

        if dtype is None:
            dtype = pxrt.getPrecision().value

        if gpu:
            assert pxd.CUPY_ENABLED
            import cupyx.scipy.sparse.linalg as spx
        else:
            spx = spsl
        return spx.LinearOperator(
            shape=self.shape,
            matvec=matmat,
            rmatvec=rmatmat,
            matmat=matmat,
            rmatmat=rmatmat,
            dtype=dtype,
        )

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        r"""
        Compute a Lipschitz constant of the operator.

        Parameters
        ----------
        method: "svd" | "trace"

            * If `svd`, compute the optimal Lipschitz constant.
            * If `trace`, compute an upper bound. (Default)

        kwargs:
            Optional kwargs passed on to:

            * `svd`: :py:func:`~pyxu.abc.LinOp.svdvals`
            * `trace`: :py:func:`~pyxu.math.hutchpp`

        Notes
        -----
        * The tightest Lipschitz constant is given by the spectral norm of the operator :math:`\mathbf{A}`:
          :math:`\|\mathbf{A}\|_{2}`.  It can be computed via the SVD, which is compute-intensive task for large
          operators.  In this setting, it may be advantageous to overestimate the Lipschitz constant with the Frobenius
          norm of :math:`\mathbf{A}` since :math:`\|\mathbf{A}\|_{F} \geq \|\mathbf{A}\|_{2}`.

          :math:`\|\mathbf{A}\|_{F}` can be efficiently approximated by computing the trace of :math:`\mathbf{A}^{\ast}
          \mathbf{A}` (or :math:`\mathbf{A}\mathbf{A}^{\ast}`) via the `Hutch++ stochastic algorithm
          <https://arxiv.org/abs/2010.09649>`_.

        * :math:`\|\mathbf{A}\|_{F}` is upper-bounded by :math:`\|\mathbf{A}\|_{F} \leq \sqrt{n} \|\mathbf{A}\|_{2}`,
          where the equality is reached (worst-case scenario) when the eigenspectrum of the linear operator is flat.
        """
        method = kwargs.get("method", "trace").lower().strip()

        if method == "svd":
            # svdvals() may have alternative signature in specialized classes, but we must always use
            # the LinOp.svdvals() interface below for kwargs-filtering.
            func, sig_func = self.__class__.svdvals, LinOp.svdvals
            kwargs.update(
                k=1,
                which="LM",
            )
            estimate = lambda: func(self, **kwargs).item()
        elif method == "trace":
            if self.shape == (1, 1):
                # [Sepand] Special case of degenerate LinOps (which are not LinFuncs).
                # Cannot call .gram() below since it will recurse indefinitely.
                # [Reason: asop().squeeze() combinations don't work for (1, 1) degenerate LinOps.]
                func = sig_func = LinFunc.estimate_lipschitz
                estimate = lambda: func(self.squeeze(), **kwargs)
            else:
                from pyxu.math import hutchpp as func

                sig_func = func

                kwargs.update(
                    op=self.gram() if (self.codim >= self.dim) else self.cogram(),
                    m=kwargs.get("m", 126),
                )
                estimate = lambda: np.sqrt(func(**kwargs)).item()
        else:
            raise NotImplementedError

        # Filter unsupported kwargs
        sig = inspect.Signature.from_callable(sig_func)
        kwargs = {k: v for (k, v) in kwargs.items() if (k in sig.parameters)}

        L = estimate()
        return L

    @pxrt.enforce_precision()
    def svdvals(
        self,
        k: pxt.Integer,
        which: str = "LM",
        gpu: bool = False,
        **kwargs,
    ) -> pxt.NDArray:
        r"""
        Compute singular values of a linear operator.

        Parameters
        ----------
        k: Integer
            Number of singular values to compute.
        which: "LM" | "SM"
            Which `k` singular values to find:

                * `LM`: largest magnitude
                * `SM`: smallest magnitude
        gpu: bool
            If ``True`` the singular value decomposition is performed on the GPU.
        kwargs: ~collections.abc.Mapping
            Additional kwargs accepted by :py:func:`~scipy.sparse.linalg.svds`.

        Returns
        -------
        D: NDArray
            (k,) singular values in ascending order.
        """
        which = which.upper().strip()
        assert which in {"SM", "LM"}

        def _dense_eval():
            if gpu:
                import cupy as xp
                import cupy.linalg as spx
            else:
                import numpy as xp
                import scipy.linalg as spx
            op = self.asarray(xp=xp, dtype=pxrt.getPrecision().value)
            return spx.svd(op, compute_uv=False)

        def _sparse_eval():
            if gpu:
                assert pxd.CUPY_ENABLED
                import cupyx.scipy.sparse.linalg as spx

                self._warn_vals_sparse_gpu()
            else:
                spx = spsl
            op = self.to_sciop(gpu=gpu, dtype=pxrt.getPrecision().value)
            kwargs.update(
                k=k,
                which=which,
                return_singular_vectors=False,
                # random_state=0,  # unsupported by CuPy
            )
            return spx.svds(op, **kwargs)

        if k >= min(self.shape) // 2:
            msg = "Too many svdvals wanted: using matrix-based ops."
            warnings.warn(msg, pxw.DenseWarning)
            D = _dense_eval()
        else:
            D = _sparse_eval()

        # Filter to k largest/smallest magnitude + sorted
        xp = pxu.get_array_module(D)
        D = D[xp.argsort(D)]
        return D[:k] if (which == "SM") else D[-k:]

    def asarray(
        self,
        xp: pxt.ArrayModule = None,
        dtype: pxt.DType = None,
    ) -> pxt.NDArray:
        r"""
        Matrix representation of the linear operator.

        Parameters
        ----------
        xp: ArrayModule
            Which array module to use to represent the output. (Default: NumPy.)
        dtype: DType
            Precision of the array. (Default: current runtime precision.)

        Returns
        -------
        A: NDArray
            (N, M) array-representation of the operator.

        Note
        ----
        This generic implementation assumes the operator is backend-agnostic.  Thus, when defining a new
        backend-specific operator, :py:meth:`~pyxu.abc.LinOp.asarray` may need to be overriden.
        """
        if xp is None:
            xp = pxd.NDArrayInfo.default().module()
        if dtype is None:
            dtype = pxrt.getPrecision().value

        with pxrt.EnforcePrecision(False):
            E = xp.eye(self.dim, dtype=dtype)
            A = self.apply(E).T
        return A

    def gram(self) -> pxt.OpT:
        r"""
        Gram operator :math:`\mathbf{A}^{\ast} \mathbf{A}: \mathbb{R}^{M} \to \mathbb{R}^{M}`.

        Returns
        -------
        op: OpT
            (M, M) Gram operator.

        Note
        ----
        By default the Gram is computed by the composition ``self.T * self``.  This may not be the fastest way to
        compute the Gram operator.  If the Gram can be computed more efficiently (e.g. with a convolution), the user
        should re-define this method.
        """

        def op_expr(_) -> tuple:
            return ("gram", self)

        op = self.T * self
        op._expr = types.MethodType(op_expr, op)
        return op.asop(SelfAdjointOp).squeeze()

    def cogram(self) -> pxt.OpT:
        r"""
        Co-Gram operator :math:`\mathbf{A}\mathbf{A}^{\ast}:\mathbb{R}^{N} \to \mathbb{R}^{N}`.

        Returns
        -------
        op: OpT
            (N, N) Co-Gram operator.

        Note
        ----
        By default the co-Gram is computed by the composition ``self * self.T``.  This may not be the fastest way to
        compute the co-Gram operator.  If the co-Gram can be computed more efficiently (e.g. with a convolution), the
        user should re-define this method.
        """

        def op_expr(_) -> tuple:
            return ("cogram", self)

        op = self * self.T
        op._expr = types.MethodType(op_expr, op)
        return op.asop(SelfAdjointOp).squeeze()

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(
        self,
        arr: pxt.NDArray,
        damp: pxt.Real,
        kwargs_init=None,
        kwargs_fit=None,
    ) -> pxt.NDArray:
        r"""
        Evaluate the Moore-Penrose pseudo-inverse :math:`\mathbf{A}^{\dagger}` at specified point(s).

        Parameters
        ----------
        arr: NDArray
            (..., N) input points.
        damp: Real
            Positive dampening factor regularizing the pseudo-inverse.
        kwargs_init: ~collections.abc.Mapping
            Optional kwargs to be passed to :py:meth:`~pyxu.opt.solver.CG`'s ``__init__()`` method.
        kwargs_fit: ~collections.abc.Mapping
            Optional kwargs to be passed to :py:meth:`~pyxu.opt.solver.CG`'s ``fit()`` method.

        Returns
        -------
        out: NDArray
            (..., M) pseudo-inverse(s).

        Notes
        -----
        The Moore-Penrose pseudo-inverse of an operator :math:`\mathbf{A}: \mathbb{R}^{M} \to \mathbb{R}^{N}` is defined
        as the operator :math:`\mathbf{A}^{\dagger}: \mathbb{R}^{N} \to \mathbb{R}^{M}` verifying the Moore-Penrose
        conditions:

            1. :math:`\mathbf{A} \mathbf{A}^{\dagger} \mathbf{A} = \mathbf{A}`,
            2. :math:`\mathbf{A}^{\dagger} \mathbf{A} \mathbf{A}^{\dagger} = \mathbf{A}^{\dagger}`,
            3. :math:`(\mathbf{A}^{\dagger} \mathbf{A})^{\ast} = \mathbf{A}^{\dagger} \mathbf{A}`,
            4. :math:`(\mathbf{A} \mathbf{A}^{\dagger})^{\ast} = \mathbf{A} \mathbf{A}^{\dagger}`.

        This operator exists and is unique for any finite-dimensional linear operator.  The action of the pseudo-inverse
        :math:`\mathbf{A}^{\dagger} \mathbf{y}` for every :math:`\mathbf{y} \in \mathbb{R}^{N}` can be computed in
        matrix-free fashion by solving the *normal equations*:

        .. math::

           \mathbf{A}^{\ast} \mathbf{A} \mathbf{x} = \mathbf{A}^{\ast} \mathbf{y}
           \quad\Leftrightarrow\quad
           \mathbf{x} = \mathbf{A}^{\dagger} \mathbf{y},
           \quad
           \forall (\mathbf{x}, \mathbf{y}) \in \mathbb{R}^{M} \times \mathbb{R}^{N}.

        In the case of severe ill-conditioning, it is possible to consider the dampened normal equations for a
        numerically-stabler approximation of :math:`\mathbf{A}^{\dagger} \mathbf{y}`:

        .. math::

           (\mathbf{A}^{\ast} \mathbf{A} + \tau I) \mathbf{x} = \mathbf{A}^{\ast} \mathbf{y},

        where :math:`\tau > 0` corresponds to the `damp` parameter.
        """
        from pyxu.operator import HomothetyOp
        from pyxu.opt.solver import CG
        from pyxu.opt.stop import MaxIter

        kwargs_fit = dict() if kwargs_fit is None else kwargs_fit
        kwargs_init = dict() if kwargs_init is None else kwargs_init
        kwargs_init.update(show_progress=kwargs_init.get("show_progress", False))

        if np.isclose(damp, 0):
            A = self.gram()
        else:
            A = self.gram() + HomothetyOp(cst=damp, dim=self.dim)

        cg = CG(A, **kwargs_init)
        if "stop_crit" not in kwargs_fit:
            # .pinv() may not have sufficiently converged given the default CG stopping criteria.
            # To avoid infinite loops, CG iterations are thresholded.
            sentinel = MaxIter(n=20 * A.dim)
            kwargs_fit["stop_crit"] = cg.default_stop_crit() | sentinel

        b = self.adjoint(arr)
        cg.fit(b=b, **kwargs_fit)
        return cg.solution()

    def dagger(
        self,
        damp: pxt.Real,
        kwargs_init=None,
        kwargs_fit=None,
    ) -> pxt.OpT:
        r"""
        Return the Moore-Penrose pseudo-inverse operator :math:`\mathbf{A}^\dagger`.

        Parameters
        ----------
        damp: Real
            Positive dampening factor regularizing the pseudo-inverse.
        kwargs_init: ~collections.abc.Mapping
            Optional kwargs to be passed to :py:meth:`~pyxu.opt.solver.CG`'s ``__init__()`` method.
        kwargs_fit: ~collections.abc.Mapping
            Optional kwargs to be passed to :py:meth:`~pyxu.opt.solver.CG`'s ``fit()`` method.

        Returns
        -------
        op: OpT
            (M, N) Moore-Penrose pseudo-inverse operator.
        """
        from pyxu.operator.interop import from_source

        def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
            return self.pinv(
                arr,
                damp=_._damp,
                kwargs_init=_._kwargs_init,
                kwargs_fit=_._kwargs_fit,
            )

        def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
            return self.T.pinv(
                arr,
                damp=_._damp,
                kwargs_init=_._kwargs_init,
                kwargs_fit=_._kwargs_fit,
            )

        kwargs_fit = dict() if kwargs_fit is None else kwargs_fit
        kwargs_init = dict() if kwargs_init is None else kwargs_init

        dagger = from_source(
            cls=SquareOp if (self.dim == self.codim) else LinOp,
            shape=(self.dim, self.codim),
            embed=dict(
                _name="dagger",
                _damp=damp,
                _kwargs_init=copy.copy(kwargs_init),
                _kwargs_fit=copy.copy(kwargs_fit),
            ),
            apply=op_apply,
            adjoint=op_adjoint,
            _expr=lambda _: (_._name, _, _._damp),
        )
        return dagger

    @classmethod
    def from_array(
        cls,
        A: typ.Union[pxt.NDArray, pxt.SparseArray],
        enable_warnings: bool = True,
    ) -> pxt.OpT:
        r"""
        Instantiate a :py:class:`~pyxu.abc.LinOp` from its array representation.

        Parameters
        ----------
        A: NDArray
            (N, M) array
        enable_warnings: bool
            If ``True``, emit a warning in case of precision mis-match issues.

        Returns
        -------
        op: OpT
            (N, M) linear operator
        """
        from pyxu.operator.linop.base import _ExplicitLinOp

        return _ExplicitLinOp(cls, A, enable_warnings)


class SquareOp(LinOp):
    r"""
    Base class for *square* linear operators, i.e. :math:`\mathbf{A}: \mathbb{R}^{M} \to \mathbb{R}^{M}`
    (endomorphsisms).
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.LINEAR_SQUARE)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        assert self.dim == self.codim, f"shape: expected (M, M), got {self.shape}."

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        """
        Compute trace of the operator.

        Parameters
        ----------
        method: "explicit" | "hutchpp"

            * If `explicit`, compute the exact trace.
            * If `hutchpp`, compute an approximation.

        kwargs: ~collections.abc.Mapping
            Optional kwargs passed to:

            * `explicit`: :py:func:`~pyxu.math.trace`
            * `hutchpp`: :py:func:`~pyxu.math.hutchpp`

        Returns
        -------
        tr: Real
            Trace estimate.
        """
        from pyxu.math import hutchpp, trace

        method = kwargs.get("method", "hutchpp").lower().strip()

        if method == "explicit":
            func = sig_func = trace
            estimate = lambda: func(op=self, **kwargs)
        elif method == "hutchpp":
            func = sig_func = hutchpp
            estimate = lambda: func(op=self, **kwargs)
        else:
            raise NotImplementedError

        # Filter unsupported kwargs
        sig = inspect.Signature.from_callable(sig_func)
        kwargs = {k: v for (k, v) in kwargs.items() if (k in sig.parameters)}

        tr = estimate()
        return tr


class NormalOp(SquareOp):
    r"""
    Base class for *normal* operators.

    Normal operators satisfy the relation :math:`\mathbf{A} \mathbf{A}^{\ast} = \mathbf{A}^{\ast} \mathbf{A}`.  It can
    be `shown <https://www.wikiwand.com/en/Spectral_theorem#/Normal_matrices>`_ that an operator is normal iff it is
    *unitarily diagonalizable*, i.e.  :math:`\mathbf{A} = \mathbf{U} \mathbf{D} \mathbf{U}^{\ast}`.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.LINEAR_NORMAL)
        return frozenset(p)

    def cogram(self) -> pxt.OpT:
        return self.gram()


class SelfAdjointOp(NormalOp):
    r"""
    Base class for *self-adjoint* operators.

    Self-adjoint operators satisfy the relation :math:`\mathbf{A}^{\ast} = \mathbf{A}`.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.LINEAR_SELF_ADJOINT)
        return frozenset(p)

    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        return self.apply(arr)


class UnitOp(NormalOp):
    r"""
    Base class for *unitary* operators.

    Unitary operators satisfy the relation :math:`\mathbf{A} \mathbf{A}^{\ast} = \mathbf{A}^{\ast} \mathbf{A} = I`.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.LINEAR_UNITARY)
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        self.lipschitz = UnitOp.estimate_lipschitz(self)

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        return 1

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        out = self.adjoint(arr)
        if not np.isclose(damp, 0):
            out = pxu.copy_if_unsafe(out)
            out /= 1 + damp
        return out

    def dagger(self, damp: pxt.Real, **kwargs) -> pxt.OpT:
        op = self.T / (1 + damp)
        return op

    def gram(self) -> pxt.OpT:
        from pyxu.operator import IdentityOp

        return IdentityOp(dim=self.dim).squeeze()

    def svdvals(self, **kwargs) -> pxt.NDArray:
        gpu = kwargs.get("gpu", False)
        xp = pxd.NDArrayInfo.from_flag(gpu).module()
        width = pxrt.getPrecision()

        D = xp.ones(kwargs["k"], dtype=width.value)
        return D


class ProjOp(SquareOp):
    r"""
    Base class for *projection* operators.

    Projection operators are *idempotent*, i.e. :math:`\mathbf{A}^{2} = \mathbf{A}`.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.LINEAR_IDEMPOTENT)
        return frozenset(p)


class OrthProjOp(ProjOp, SelfAdjointOp):
    r"""
    Base class for *orthogonal projection* operators.

    Orthogonal projection operators are *idempotent* and *self-adjoint*, i.e.  :math:`\mathbf{A}^{2} = \mathbf{A}` and
    :math:`\mathbf{A}^{\ast} = \mathbf{A}`.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set()
        for klass in cls.__bases__:
            p |= klass.properties()
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        self.lipschitz = OrthProjOp.estimate_lipschitz(self)

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        return 1

    def gram(self) -> pxt.OpT:
        return self.squeeze()

    def cogram(self) -> pxt.OpT:
        return self.squeeze()

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        out = self.apply(arr)
        if not np.isclose(damp, 0):
            out = pxu.copy_if_unsafe(out)
            out /= 1 + damp
        return out

    def dagger(self, damp: pxt.Real, **kwargs) -> pxt.OpT:
        op = self / (1 + damp)
        return op


class PosDefOp(SelfAdjointOp):
    r"""
    Base class for *positive-definite* operators.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set(super().properties())
        p.add(Property.LINEAR_POSITIVE_DEFINITE)
        return frozenset(p)


class LinFunc(ProxDiffFunc, LinOp):
    r"""
    Base class for real-valued linear functionals :math:`f: \mathbb{R}^{M} \to \mathbb{R}`.

    Instances of this class must implement :py:meth:`~pyxu.abc.Map.apply`, and :py:meth:`~pyxu.abc.LinOp.adjoint`.

    If known, the Lipschitz constant :math:`L` should be stored in the :py:attr:`~pyxu.abc.Map.lipschitz` property.
    """

    @classmethod
    def properties(cls) -> cabc.Set[Property]:
        p = set()
        for klass in cls.__bases__:
            p |= klass.properties()
        return frozenset(p)

    def __init__(self, shape: pxt.OpShape):
        super().__init__(shape=shape)
        ProxDiffFunc.__init__(self, shape)
        LinOp.__init__(self, shape)

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        return LinOp.jacobian(self, arr)

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        # Try all backends until one works.
        for ndi in pxd.NDArrayInfo:
            try:
                xp = ndi.module()
                g = self.grad(xp.ones(self.dim))
                L = float(xp.linalg.norm(g))
                return L
            except Exception:
                pass

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        x = xp.ones((*arr.shape[:-1], 1), dtype=arr.dtype)
        g = self.adjoint(x)
        return g

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        # out = arr - tau * self.grad(arr)
        out = pxu.copy_if_unsafe(self.grad(arr))
        out *= -tau
        out += arr
        return out

    @pxrt.enforce_precision(i=("arr", "sigma"))
    def fenchel_prox(self, arr: pxt.NDArray, sigma: pxt.Real) -> pxt.NDArray:
        return self.grad(arr)

    def cogram(self) -> pxt.OpT:
        from pyxu.operator import HomothetyOp

        L = self.estimate_lipschitz()
        return HomothetyOp(cst=L**2, dim=1)

    def svdvals(self, **kwargs) -> pxt.NDArray:
        gpu = kwargs.get("gpu", False)
        xp = pxd.NDArrayInfo.from_flag(gpu).module()
        width = pxrt.getPrecision()

        L = self.estimate_lipschitz()
        D = xp.array([L], dtype=width.value)
        return D

    def asarray(self, **kwargs) -> pxt.NDArray:
        xp = kwargs.get("xp", pxd.NDArrayInfo.default().module())
        dtype = kwargs.get("dtype", pxrt.getPrecision().value)

        with pxrt.EnforcePrecision(False):
            x = xp.ones((1, 1), dtype=dtype)
            A = self.adjoint(x)
        return A

    @classmethod
    def from_array(
        cls,
        A: typ.Union[pxt.NDArray, pxt.SparseArray],
        enable_warnings: bool = True,
    ) -> pxt.OpT:
        if A.ndim == 1:
            A = A.reshape((1, -1))
        op = super().from_array(A, enable_warnings)
        return op


def _core_operators() -> cabc.Set[pxt.OpC]:
    # Operators which can be sub-classed by end-users and participate in arithmetic rules.
    ops = set()
    for _ in globals().values():
        if inspect.isclass(_) and issubclass(_, Operator):
            ops.add(_)
    ops.remove(Operator)
    return ops


def _is_real(x) -> bool:
    if isinstance(x, pxt.Real):
        return True
    elif isinstance(x, pxd.supported_array_types()) and (x.size == 1):
        return True
    else:
        return False


__all__ = [
    "Operator",
    "Property",
    *map(lambda _: _.__name__, _core_operators()),
]



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class TestNullOp(conftest.LinOpT):
    @pytest.fixture
    def data_shape(self):
        return (3, 4)

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, data_shape, request):
        op = pxo.NullOp(shape=data_shape)
        return op, *request.param

    @pytest.fixture
    def data_apply(self, data_shape):
        arr = np.arange(data_shape[1])
        out = np.zeros(data_shape[0])
        return dict(
            in_=dict(arr=arr),
            out=out,
        )



import numpy as np
import pytest

import pyxu.info.ptype as pxt
import pyxu.math as pxm
import pyxu.runtime as pxrt
import pyxu_tests.conftest as ct


def op_squareop():
    import pyxu_tests.operator.examples.test_squareop as tc

    return tc.CumSum(N=7)


def op_normalop():
    import pyxu_tests.operator.examples.test_normalop as tc

    rng = np.random.default_rng(seed=2)
    h = rng.normal(size=(7,))
    return tc.CircularConvolution(h=h)


def op_unitop():
    import pyxu_tests.operator.examples.test_unitop as tc

    return tc.Permutation(N=7)


def op_selfadjointop():
    import pyxu_tests.operator.examples.test_selfadjointop as tc

    return tc.SelfAdjointConvolution(N=7)


def op_posdefop():
    import pyxu_tests.operator.examples.test_posdefop as tc

    return tc.PSDConvolution(N=7)


def op_projop():
    import pyxu_tests.operator.examples.test_projop as tc

    return tc.Oblique(N=7, alpha=np.pi / 4)


def op_orthprojop():
    import pyxu_tests.operator.examples.test_orthprojop as tc

    return tc.ScaleDown(N=7)


class TestTrace:
    @pytest.fixture(
        params=[
            op_squareop(),
            op_normalop(),
            op_unitop(),
            op_selfadjointop(),
            op_posdefop(),
            op_projop(),
            op_orthprojop(),
        ]
    )
    def op(self, request) -> pxt.OpT:
        return request.param

    @pytest.fixture
    def _op_trace(self, op) -> float:
        # Ground truth trace
        tr = op.asarray().trace()
        return tr

    def test_value_explicit(self, op, _op_trace):
        tr = pxm.trace(op)
        assert ct.allclose(tr, _op_trace, as_dtype=pxrt.getPrecision().value)

    def test_value_hutchpp(self, op, _op_trace):
        # Ensure computed trace (w/ default parameter values) satisfies statistical property stated
        # in hutchpp() docstring, i.e.: estimation error smaller than 1e-2 w/ probability 0.9
        N_trial = 100
        tr = np.array([pxm.hutchpp(op) for _ in range(N_trial)])
        N_pass = sum(np.abs(tr - _op_trace) <= 1e-2)
        assert N_pass >= 0.9 * N_trial



import collections.abc as cabc
import typing as typ

import numpy as np

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "Pad",
]


class Pad(pxa.LinOp):
    r"""
    Multi-dimensional padding operator.

    This operator pads the input array in each dimension according to specified widths.

    Notes
    -----
    * If inputs are D-dimensional, then some of the padding of later axes are calculated from padding of previous axes.
    * The *adjoint* of the padding operator performs a cumulative summation over the original positions used to pad.
      Its effect is clear from its matrix form.  For example the matrix-form of ``Pad(arg_shape=(3,), mode="wrap",
      pad_width=(1, 1))`` is:

      .. math::

         \mathbf{A}
         =
         \left[
            \begin{array}{ccc}
                0 & 0 & 1 \\
                1 & 0 & 0 \\
                0 & 1 & 0 \\
                0 & 0 & 1 \\
                1 & 0 & 0
            \end{array}
         \right].

      The adjoint of :math:`\mathbf{A}` corresponds to its matrix transpose:

      .. math::

         \mathbf{A}^{\ast}
         =
         \left[
             \begin{array}{ccccc}
                 0 & 1 & 0 & 0 & 1 \\
                 0 & 0 & 1 & 0 & 0 \\
                 1 & 0 & 0 & 1 & 0
             \end{array}
         \right].

      This operation can be seen as a trimming (:math:`\mathbf{T}`) plus a cumulative summation (:math:`\mathbf{S}`):

      .. math::

         \mathbf{A}^{\ast}
         =
         \mathbf{T} + \mathbf{S}
         =
         \left[
            \begin{array}{ccccc}
                0 & 1 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1 & 0
            \end{array}
         \right]
         +
         \left[
            \begin{array}{ccccc}
                0 & 0 & 0 & 0 & 1 \\
                0 & 0 & 0 & 0 & 0 \\
                1 & 0 & 0 & 0 & 0
            \end{array}
         \right],

      where both :math:`\mathbf{T}` and :math:`\mathbf{S}` are efficiently implemented in matrix-free form.


    * The Lipschitz constant of the multi-dimensional padding operator is upper-bounded by the product of Lipschitz
      constants of the uni-dimensional paddings applied per dimension, i.e.:

      .. math::

         L \le \prod_{i} L_{i}, \qquad i \in \{0, \dots, N-1\},

      where :math:`L_{i}` depends on the boundary condition at the :math:`i`-th axis.

      :math:`L_{i}^{2}` corresponds to the maximum singular value of the diagonal matrix

      .. math::

         \mathbf{A}_{i}^{\ast} \mathbf{A}_{i}
         =
         \mathbf{T}_{i}^{\ast} \mathbf{T}_{i} + \mathbf{S}_{i}^{\ast} \mathbf{S}_{i}
         =
         \mathbf{I}_{N} + \mathbf{S}_{i}^{\ast} \mathbf{S}_{i}.

      - In mode="constant", :math:`\text{diag}(\mathbf{S}_{i}^{\ast} \mathbf{S}_{i}) = \mathbf{0}`, hence :math:`L_{i} =
        1`.
      - In mode="edge",

        .. math::

           \text{diag}(\mathbf{S}_{i}^{\ast} \mathbf{S}_{i})
           =
           \left[p_{lhs}, 0, \ldots, 0, p_{rhs} \right],

        hence :math:`L_{i} = \sqrt{1 + \min(p_{lhs}, p_{rhs})}`.
      - In mode="symmetric", "wrap", "reflect", :math:`\text{diag}(\mathbf{S}_{i}^{\ast} \mathbf{S}_{i})` equals (up to
        a mode-dependant permutation)

        .. math::

           \text{diag}(\mathbf{S}_{i}^{\ast} \mathbf{S}_{i})
           =
           \left[1, \ldots, 1, 0, \ldots, 0\right]
           +
           \left[0, \ldots, 0, 1, \ldots, 1\right],

        hence

        .. math::

           L^{\text{wrap, symmetric}}_{i} = \sqrt{1 + \lceil\frac{p_{lhs} + p_{rhs}}{N}\rceil}, \\
           L^{\text{reflect}}_{i} = \sqrt{1 + \lceil\frac{p_{lhs} + p_{rhs}}{N-2}\rceil}.
    """
    WidthSpec = typ.Union[
        pxt.Integer,
        cabc.Sequence[pxt.Integer],
        cabc.Sequence[tuple[pxt.Integer, pxt.Integer]],
    ]
    ModeSpec = typ.Union[str, cabc.Sequence[str]]

    def __init__(
        self,
        arg_shape: pxt.NDArrayShape,
        pad_width: WidthSpec,
        mode: ModeSpec = "constant",
    ):
        r"""
        Parameters
        ----------
        arg_shape: NDArrayShape
            Shape of the input array.
        pad_width: ~pyxu.operator.linop.pad.Pad.WidthSpec
            Number of values padded to the edges of each axis.
            Multiple forms are accepted:

            * ``int``: pad each dimension's head/tail by `pad_width`.
            * ``tuple[int, ...]``: pad dimension[k]'s head/tail by `pad_width[k]`.
            * ``tuple[tuple[int, int], ...]``: pad dimension[k]'s head/tail by `pad_width[k][0]` /
              `pad_width[k][1]` respectively.
        mode: str, :py:class:`list` ( str )
            Padding mode.
            Multiple forms are accepted:

            * str: unique mode shared amongst dimensions.
              Must be one of:

              * 'constant' (zero-padding)
              * 'wrap'
              * 'reflect'
              * 'symmetric'
              * 'edge'
            * tuple[str, ...]: pad dimension[k] using `mode[k]`.

            (See :py:func:`numpy.pad` for details.)
        """
        self._arg_shape = tuple(arg_shape)
        assert all(_ > 0 for _ in self._arg_shape)
        N_dim = len(self._arg_shape)

        # transform `pad_width` to canonical form tuple[tuple[int, int], ...]
        is_seq = lambda _: isinstance(_, cabc.Sequence)
        if not is_seq(pad_width):  # int-form
            pad_width = ((pad_width, pad_width),) * N_dim
        assert len(pad_width) == N_dim, "arg_shape/pad_width are length-mismatched."
        if not is_seq(pad_width[0]):  # tuple[int, ...] form
            pad_width = tuple((w, w) for w in pad_width)
        else:  # tuple[tulpe[int, int], ...] form
            pass
        assert all(0 <= min(lhs, rhs) for (lhs, rhs) in pad_width)
        self._pad_width = tuple(pad_width)

        # transform `mode` to canonical form tuple[str, ...]
        if isinstance(mode, str):  # shared mode
            mode = (mode,) * N_dim
        elif isinstance(mode, cabc.Sequence):  # tuple[str, ...]: different modes
            assert len(mode) == N_dim, "arg_shape/mode are length-mismatched."
            mode = tuple(mode)
        else:
            raise ValueError(f"Unkwown mode encountered: {mode}.")
        self._mode = tuple(map(lambda _: _.strip().lower(), mode))
        assert set(self._mode) <= {
            "constant",
            "wrap",
            "reflect",
            "symmetric",
            "edge",
        }, "Unknown mode(s) encountered."

        # Useful constant: shape of padded outputs
        self._pad_shape = list(self._arg_shape)
        for i, (lhs, rhs) in enumerate(self._pad_width):
            self._pad_shape[i] += lhs + rhs
        self._pad_shape = tuple(self._pad_shape)

        codim = np.prod(self._pad_shape)
        dim = np.prod(self._arg_shape)
        super().__init__(shape=(codim, dim))

        # Some modes have awkward interpretations when pad-widths cross certain thresholds.
        # Supported pad-widths are thus limited to sensible regions.
        for i in range(N_dim):
            N = self._arg_shape[i]
            w_max = dict(
                constant=np.inf,
                wrap=N,
                reflect=N - 1,
                symmetric=N,
                edge=np.inf,
            )[self._mode[i]]
            lhs, rhs = self._pad_width[i]
            assert max(lhs, rhs) <= w_max, f"pad_width along dim-{i} is limited to {w_max}."

        # We know a crude Lipschitz bound by default. Since computing it takes (code) space,
        # the estimate is computed as a special case of estimate_lipschitz()
        self.lipschitz = self.estimate_lipschitz(__rule=True)

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        sh = arr.shape[:-1]
        arr = arr.reshape(*sh, *self._arg_shape)
        N_dim = len(self._arg_shape)

        # Part 1: extend the core
        xp = pxu.get_array_module(arr)
        pad_width_sh = ((0, 0),) * len(sh)  # don't pad stack-dims
        out = xp.pad(
            array=arr,
            pad_width=pad_width_sh + self._pad_width,
            mode="constant",
            constant_values=0,
        )

        # Part 2: apply border effects (if any)
        for i in range(N_dim, 0, -1):
            mode = self._mode[-i]
            lhs, rhs = self._pad_width[-i]
            N = self._pad_shape[-i]

            r_s = [slice(None)] * (len(sh) + N_dim)  # read axial selector
            w_s = [slice(None)] * (len(sh) + N_dim)  # write axial selector

            if mode == "constant":
                # no border effects
                pass
            elif mode == "wrap":
                # Fix LHS
                r_s[-i] = slice(N - rhs - lhs, N - rhs)
                w_s[-i] = slice(0, lhs)
                out[tuple(w_s)] = out[tuple(r_s)]

                # Fix RHS
                r_s[-i] = slice(lhs, lhs + rhs)
                w_s[-i] = slice(N - rhs, N)
                out[tuple(w_s)] = out[tuple(r_s)]
            elif mode == "reflect":
                # Fix LHS
                r_s[-i] = slice(2 * lhs, lhs, -1)
                w_s[-i] = slice(0, lhs)
                out[tuple(w_s)] = out[tuple(r_s)]

                # Fix RHS
                r_s[-i] = slice(N - rhs - 2, N - 2 * rhs - 2, -1)
                w_s[-i] = slice(N - rhs, N)
                out[tuple(w_s)] = out[tuple(r_s)]
            elif mode == "symmetric":
                # Fix LHS
                r_s[-i] = slice(2 * lhs - 1, lhs - 1, -1)
                w_s[-i] = slice(0, lhs)
                out[tuple(w_s)] = out[tuple(r_s)]

                # Fix RHS
                r_s[-i] = slice(N - rhs - 1, N - 2 * rhs - 1, -1)
                w_s[-i] = slice(N - rhs, N)
                out[tuple(w_s)] = out[tuple(r_s)]
            elif mode == "edge":
                if lhs > 0:  # Fix LHS
                    r_s[-i] = slice(lhs, lhs + 1)
                    w_s[-i] = slice(0, lhs)
                    out[tuple(w_s)] = out[tuple(r_s)]

                if rhs > 0:  # Fix RHS
                    r_s[-i] = slice(N - rhs - 1, N - rhs)
                    w_s[-i] = slice(N - rhs, N)
                    out[tuple(w_s)] = out[tuple(r_s)]

        out = out.reshape(*sh, self.codim)
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        sh = arr.shape[:-1]
        arr = arr.reshape(*sh, *self._pad_shape)
        N_dim = len(self._arg_shape)

        # Part 1: apply correction terms (if any)
        out = arr.copy()  # in-place updates below
        for i in range(1, N_dim + 1):
            mode = self._mode[-i]
            lhs, rhs = self._pad_width[-i]
            N = self._pad_shape[-i]

            r_s = [slice(None)] * (len(sh) + N_dim)  # read axial selector
            w_s = [slice(None)] * (len(sh) + N_dim)  # write axial selector

            if mode == "constant":
                # no correction required
                pass
            elif mode == "wrap":
                # Fix LHS
                r_s[-i] = slice(0, lhs)
                w_s[-i] = slice(N - rhs - lhs, N - rhs)
                out[tuple(w_s)] += out[tuple(r_s)]

                # Fix RHS
                r_s[-i] = slice(N - rhs, N)
                w_s[-i] = slice(lhs, lhs + rhs)
                out[tuple(w_s)] += out[tuple(r_s)]
            elif mode == "reflect":
                if lhs > 0:  # Fix LHS
                    r_s[-i] = slice(lhs - 1, None, -1)
                    w_s[-i] = slice(lhs + 1, 2 * lhs + 1)
                    out[tuple(w_s)] += out[tuple(r_s)]

                # Fix RHS
                r_s[-i] = slice(N - 1, N - rhs - 1, -1)
                w_s[-i] = slice(N - 2 * rhs - 1, N - rhs - 1)
                out[tuple(w_s)] += out[tuple(r_s)]
            elif mode == "symmetric":
                if lhs > 0:  # Fix LHS
                    r_s[-i] = slice(lhs - 1, None, -1)
                    w_s[-i] = slice(lhs, 2 * lhs)
                    out[tuple(w_s)] += out[tuple(r_s)]

                # Fix RHS
                r_s[-i] = slice(N - 1, N - rhs - 1, -1)
                w_s[-i] = slice(N - 2 * rhs, N - rhs)
                out[tuple(w_s)] += out[tuple(r_s)]
            elif mode == "edge":
                if lhs > 0:  # Fix LHS
                    r_s[-i] = slice(0, lhs)
                    w_s[-i] = slice(lhs, lhs + 1)
                    out[tuple(w_s)] += out[tuple(r_s)].sum(axis=-i, keepdims=True)

                if rhs > 0:  # Fix RHS
                    r_s[-i] = slice(N - rhs, N)
                    w_s[-i] = slice(N - rhs - 1, N - rhs)
                    out[tuple(w_s)] += out[tuple(r_s)].sum(axis=-i, keepdims=True)

        # Part 2: extract the core
        selector = [slice(None)] * len(sh)
        for N, (lhs, rhs) in zip(self._pad_shape, self._pad_width):
            s = slice(lhs, N - rhs)
            selector.append(s)
        out = out[tuple(selector)]

        out = out.reshape(*sh, self.dim)
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L = []  # 1D pad-op Lipschitz constants
            for N, m, (lhs, rhs) in zip(self._arg_shape, self._mode, self._pad_width):
                if m == "constant":
                    _L = 1
                elif m in {"wrap", "symmetric"}:
                    _L = np.sqrt(1 + np.ceil((lhs + rhs) / N))
                elif m == "reflect":
                    _L = np.sqrt(1 + np.ceil((lhs + rhs) / (N - 2)))
                elif m == "edge":
                    _L = np.sqrt(1 + max(lhs, rhs))
                L.append(_L)
            L = np.prod(L)
        else:
            L = super().estimate_lipschitz(**kwargs)
        return L

    def gram(self) -> pxt.OpT:
        if all(m == "constant" for m in self._mode):
            from pyxu.operator.linop.base import IdentityOp

            op = IdentityOp(dim=self.dim)
        else:
            op = super().gram()
        return op

    def cogram(self) -> pxt.OpT:
        if all(m == "constant" for m in self._mode):
            from pyxu.operator.linop.select import Trim

            # Orthogonal projection
            op = Trim(
                arg_shape=self._pad_shape,
                trim_width=self._pad_width,
            ).gram()
        else:
            op = super().cogram()
        return op



r"""
This sampler module implements state-of-the-art algorithms that generate samples from probability distributions.  These
algorithms are particularly well-suited for high-dimensional distributions such as posterior distributions of inverse
problems in imaging, which is a notoriously difficult task.  The ability to sample from the posterior distribution is
extremely valuable, as it allows to explore the landscape of the objective function instead of having a single point
estimate (the maximum a posteriori solution, i.e. the mode of the posterior).  This is useful for uncertainty
quantification (UQ) purposes [UQ_MCMC]_, and it allows compute Monte Carlo estimates of expected values with respect to
the posterior.  For example, the mean of samples from the posterior is an approximation of the minimum mean-square error
(MMSE) estimator that can be used for image reconstruction.  Higher-order pixel-wise statistics (e.g., the variance) can
also be computed in an online fashion (see :py:mod:`~pyxu.experimental.sampler.statistics`) and provide useful
diagnostic tools for uncertainty quantification.

In the following example, we showcase the unajusted Langevin algorithm (:py:class:`~pyxu.experimental.sampler.ULA`)
applied to a total-variation denoising problem.  We show the MMSE estimator as well as the pixelwise variance of the
samples.  As expected, the variance is higher around edges than in the smooth regions, indicating that there is higher
uncertainty in these regions.

.. code-block:: python3

   import matplotlib.pyplot as plt
   import numpy as np
   import pyxu.experimental.sampler as pxe_sampler
   import pyxu.operator as pxo
   import pyxu.opt.solver as pxsl
   import skimage as skim

   sh_im = (128,) * 2
   gt = skim.transform.resize(skim.data.shepp_logan_phantom(), sh_im)  # Ground-truth image
   N = np.prod(sh_im)  # Number of pixels

   # Noisy data
   rng = np.random.default_rng(seed=0)
   sigma = 1e-1  # Noise standard deviation
   y = gt + sigma * rng.standard_normal(sh_im)  # Noisy image
   f = 1 / 2 * pxo.SquaredL2Norm(dim=N).argshift(-y.ravel()) / sigma**2  # Data fidelity loss

   # Smoothed TV regularization
   g = pxo.L21Norm(arg_shape=(2, *sh_im)).moreau_envelope(1e-2) * pxo.Gradient(arg_shape=sh_im)
   theta = 10  # Regularization parameter

   # Compute MAP estimator
   pgd = pxsl.PGD(f=f + theta * g)
   pgd.fit(x0=y.ravel())
   im_MAP = pgd.solution().reshape(sh_im)

   fig, ax = plt.subplots(1, 3)
   ax[0].imshow(gt)
   ax[0].set_title("Ground truth")
   ax[0].axis("off")
   ax[1].imshow(y, vmin=0, vmax=1)
   ax[1].set_title("Noisy image")
   ax[1].axis("off")
   ax[2].imshow(im_MAP, vmin=0, vmax=1)
   ax[2].set_title("MAP reconstruction")
   ax[2].axis("off")

   ula = pxe_sampler.ULA(f=f + theta * g)  # ULA sampler

   n = int(1e4)  # Number of samples
   burn_in = int(1e3)  # Number of burn-in iterations
   gen = ula.samples(x0=np.zeros(N), rng=rng)  # Generator for ULA samples
   # Objects for computing online statistics based on samples
   online_mean = pxe_sampler.OnlineMoment(order=1)
   online_var = pxe_sampler.OnlineVariance()

   i = 0  # Number of samples
   for sample in gen:  # Draw ULA sample
       i += 1
       if i > burn_in + n:
           break
       if i > burn_in:
           mean = online_mean.update(sample)  # Update online mean
           var = online_var.update(sample)  # Update online variance

   fig, ax = plt.subplots(1, 2)
   mean_im = ax[0].imshow(mean.reshape(sh_im), vmin=0, vmax=1)
   fig.colorbar(mean_im, fraction=0.05, ax=ax[0])
   ax[0].set_title("Mean (MMSE estimator)")
   ax[0].axis("off")
   var_im = ax[1].imshow(var.reshape(sh_im))
   fig.colorbar(var_im, fraction=0.05, ax=ax[1])
   ax[1].set_title("Variance")
   ax[1].axis("off")
   fig.suptitle("Pixel-wise statistics of ULA samples")
"""

import collections.abc as cabc
import math

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "_Sampler",
    "ULA",
    "MYULA",
]


class _Sampler:
    """Abstract base class for samplers."""

    def samples(self, rng=None, **kwargs) -> cabc.Generator:
        """Returns a generator; samples are drawn by calling next(generator)."""
        self._sample_init(rng, **kwargs)

        def _generator():
            while True:
                yield self._sample()

        return _generator()

    def _sample_init(self, rng, **kwargs):
        """Optional method to set initial state of the sampler (e.g., a starting point)."""
        pass

    def _sample(self) -> pxt.NDArray:
        """Method to be implemented by subclasses that returns the next sample."""
        raise NotImplementedError


class ULA(_Sampler):
    r"""
    Unajusted Langevin algorithm (ULA).

    Generates samples from the distribution

    .. math::

       p(\mathbf{x})
       =
       \frac{\exp(-\mathcal{F}(\mathbf{x}))}{\int_{\mathbb{R}^N} \exp(-\mathcal{F}(\tilde{\mathbf{x}}))
       \mathrm{d} \tilde{\mathbf{x}} },

    where :math:`\mathcal{F}: \mathbb{R}^N \to \mathbb{R}` is *differentiable* with :math:`\beta`-*Lipschitz continuous*
    gradient.

    Notes
    -----
    ULA is a Monte-Carlo Markov chain (MCMC) method that derives from the discretization of overdamped Langevin
    diffusions.  More specifically, it relies on the Langevin stochastic differential equation (SDE):

    .. math::

       \mathrm{d} \mathbf{X}_t
       =
       - \nabla \mathcal{F}(\mathbf{X}_t) \mathrm{d}t + \sqrt{2} \mathrm{d} \mathbf{B}_t,

    where :math:`(\mathbf{B}_t)_{t \geq 0}` is a :math:`N`-dimensional Brownian motion.  It is well known that under
    mild technical assumptions, this SDE has a unique strong solution whose invariant distribution is
    :math:`p(\mathbf{x}) \propto \exp(-\mathcal{F}(\mathbf{x}))`.  The discrete-time Euler-Maruyama discretization of
    this SDE then yields the ULA Markov chain

    .. math::

       \mathbf{X}_{k+1} = \mathbf{X}_{k} - \gamma \nabla \mathcal{F}(\mathbf{X}_k) + \sqrt{2 \gamma} \mathbf{Z}_{k+1}

    for all :math:`k \in \mathbb{Z}`, where :math:`\gamma` is the discretization step size and :math:`(\mathbf{Z}_k)_{k
    \in \mathbb{Z}}` is a sequence of independant and identically distributed :math:`N`-dimensional standard Gaussian
    distributions.  When :math:`\mathcal{F}` is differentiable with :math:`\beta`-Lipschitz continuous gradient and
    :math:`\gamma \leq \frac{1}{\beta}`, the ULA Markov chain converges (see [ULA]_) to a unique stationary distribution
    :math:`p_\gamma` such that

    .. math::

       \lim_{\gamma \to 0} \Vert p_\gamma - p \Vert_{\mathrm{TV}} = 0.

    The discretization step :math:`\gamma` is subject to the bias-variance tradeoff: a larger step will lead to faster
    convergence of the Markov chain at the expense of a larger bias in the approximation of the distribution :math:`p`.
    Setting :math:`\gamma` as large as possible (default behavior) is recommended for large-scale problems, since
    convergence speed (rather than approximation bias) is then typically the main bottelneck.  See `Example` section
    below for a concrete illustration of this tradeoff.

    Remarks
    -------
    Like all MCMC sampling methods, ULA comes with the following challenges:

    * The first few samples of the chain may not be adequate for computing statistics, as they might be located in low
      probability regions.  This challenge can either be alleviated by selecting a representative starting point to the
      chain, or by having a `burn-in` phase where the first few samples are discarded.

    * Consecutive samples are typically correlated, which can deteriorate the Monte-Carlo estimation of quantities of
      interest.  This issue can be alleviated by `thinning` the chain, i.e., selecting only every :math:`k` samples, at
      the expense of an increased computational cost.  Useful diagnostic tools to quantify this correlation between
      samples include the pixel-wise autocorrelation function and the `effective sample size
      <https://mc-stan.org/docs/reference-manual/effective-sample-size.html>`_.

    Example
    -------
    We illustrate ULA on a 1D example (:math:`N = 1`) where :math:`\mathcal{F}(x) = \frac{x^2}{2}`; the target
    distribution :math:`p(x)` is thus the 1D standard Gaussian.  In this toy example, the biased distribution
    :math:`p_\gamma(x)` can be computed in closed form.  The ULA Markov chain is given by

    .. math::

       \mathbf{X}_{k+1} &= \mathbf{X}_{k} - \gamma \nabla\mathcal{F}(\mathbf{X}_k) + \sqrt{2\gamma}\mathbf{Z}_{k+1} \\
       &= \mathbf{X}_{k} (1 - \gamma) + \sqrt{2 \gamma} \mathbf{Z}_{k+1}.

    Assuming for simplicity that :math:`\mathbf{X}_0` is Gaussian with mean :math:`\mu_0` and variance
    :math:`\sigma_0^2`, :math:`\mathbf{X}_k` is Gaussian for any :math:`k \in \mathbb{Z}` as a linear combination of
    Gaussians.  Taking the expected value of the recurrence relation yields

    .. math::

       \mu_k := \mathbb{E}(\mathbf{X}_{k}) = \mathbb{E}(\mathbf{X}_{k-1}) (1 - \gamma) = \mu_0 (1 - \gamma)^k

    (geometric sequence).  Taking the expected value of the square of the recurrence relation yields

    .. math::

       \mu^{(2)}_k := \mathbb{E}(\mathbf{X}_{k}^2) = \mathbb{E}(\mathbf{X}_{k-1}^2) (1 - \gamma)^2 + 2 \gamma =
       (1 - \gamma)^{2k} (\sigma_0^2 - b) + b

    with :math:`b = \frac{2 \gamma}{1 - (1 - \gamma)^{2}} = \frac{1}{1-\frac{\gamma}{2}}` (arithmetico-geometric
    sequence) due to the independence of :math:`\mathbf{X}_{k-1}` and :math:`\mathbf{Z}_{k}`.  Hence,
    :math:`p_\gamma(x)` is a Gaussian with mean :math:`\mu_\gamma= \lim_{k \to \infty} \mu_k = 0` and variance
    :math:`\sigma_\gamma^2 = \lim_{k \to \infty} \mu^{(2)}_k - \mu_k^2 = \frac{1}{1-\frac{\gamma}{2}}`.  As expected, we
    have :math:`\lim_{\gamma \to 0} \sigma_\gamma^2 = 1`, which is the variance of the target distribution :math:`p(x)`.

    We plot the distribution of the samples of ULA for one large (:math:`\gamma_1 \approx 1`, i.e.
    :math:`\sigma_{\gamma_1}^2 \approx 2`) and one small (:math:`\gamma_2 = 0.1`, i.e. :math:`\sigma_{\gamma_2}^2
    \approx 1.05`) step size.  As expected, the larger step size :math:`\gamma_1` leads to a larger bias in the
    approximation of :math:`p(x)`.  To quantify the speed of convergence of the Markov chains, we compute the
    `Cramér-von Mises <https://en.wikipedia.org/wiki/Cram%C3%A9r%E2%80%93von_Mises_criterion>`_ tests of goodness of fit
    of the empirical distributions to the stationary distributions of ULA :math:`p_{\gamma_1}(x)` and
    :math:`p_{\gamma_2}(x)`.  We observe that the larger step :math:`\gamma_1` leads to a better fit (lower Cramér-von
    Mises criterion), which illustrates the aforementioned bias-variance tradeoff for the choice of the step size.

    .. plot::

       import matplotlib.pyplot as plt
       import numpy as np
       import pyxu.experimental.sampler as pxe_sampler
       import pyxu.operator as pxo
       import scipy as sp

       f = pxo.SquaredL2Norm(dim=1) / 2  # To sample 1D normal distribution (mean 0, variance 1)
       ula = pxe_sampler.ULA(f=f)  # Sampler with maximum step size
       ula_lb = pxe_sampler.ULA(f=f, gamma=1e-1)  # Sampler with small step size

       gen_ula = ula.samples(x0=np.zeros(1))
       gen_ula_lb = ula_lb.samples(x0=np.zeros(1))
       n_burn_in = int(1e3)  # Number of burn-in iterations
       for i in range(n_burn_in):
           next(gen_ula)
           next(gen_ula_lb)

       # Online statistics objects
       mean_ula = pxe_sampler.OnlineMoment(order=1)
       mean_ula_lb = pxe_sampler.OnlineMoment(order=1)
       var_ula = pxe_sampler.OnlineVariance()
       var_ula_lb = pxe_sampler.OnlineVariance()

       n = int(1e4)  # Number of samples
       samples_ula = np.zeros(n)
       samples_ula_lb = np.zeros(n)
       for i in range(n):
           sample = next(gen_ula)
           sample_lb = next(gen_ula_lb)
           samples_ula[i] = sample
           samples_ula_lb[i] = sample_lb
           mean = float(mean_ula.update(sample))
           var = float(var_ula.update(sample))
           mean_lb = float(mean_ula_lb.update(sample_lb))
           var_lb = float(var_ula_lb.update(sample_lb))

       # Theoretical variances of biased stationary distributions of ULA
       biased_var = 1 / (1 - ula._gamma / 2)
       biased_var_lb = 1 / (1 - ula_lb._gamma / 2)

       # Quantify goodness of fit of empirical distribution with theoretical distribution (Cramér-von Mises test)
       cvm = sp.stats.cramervonmises(samples_ula, "norm", args=(0, np.sqrt(biased_var)))
       cvm_lb = sp.stats.cramervonmises(samples_ula_lb, "norm", args=(0, np.sqrt(biased_var_lb)))

       # Plots
       grid = np.linspace(-4, 4, 1000)

       plt.figure()
       plt.title(
           f"ULA samples (large step size) \n Empirical mean: {mean:.3f} (theoretical: 0) \n "
           f"Empirical variance: {var:.3f} (theoretical: {biased_var:.3f}) \n"
           f"Cramér-von Mises goodness of fit: {cvm.statistic:.3f}"
       )
       plt.hist(samples_ula, range=(min(grid), max(grid)), bins=100, density=True)
       plt.plot(grid, sp.stats.norm.pdf(grid), label=r"$p(x)$")
       plt.plot(grid, sp.stats.norm.pdf(grid, scale=np.sqrt(biased_var)), label=r"$p_{\gamma_1}(x)$")
       plt.legend()
       plt.show()

       plt.figure()
       plt.title(
           f"ULA samples (small step size) \n Empirical mean: {mean_lb:.3f} (theoretical: 0) \n "
           f"Empirical variance: {var_lb:.3f} (theoretical: {biased_var_lb:.3f}) \n"
           f"Cramér-von Mises goodness of fit: {cvm_lb.statistic:.3f}"
       )
       plt.hist(samples_ula_lb, range=(min(grid), max(grid)), bins=100, density=True)
       plt.plot(grid, sp.stats.norm.pdf(grid), label=r"$p(x)$")
       plt.plot(grid, sp.stats.norm.pdf(grid, scale=np.sqrt(biased_var_lb)), label=r"$p_{\gamma_2}(x)$")
       plt.legend()
       plt.show()
    """

    def __init__(self, f: pxa.DiffFunc, gamma: pxt.Real = None):
        r"""
        Parameters
        ----------
        f: :py:class:`~pyxu.abc.DiffFunc`
            Differentiable functional.
        gamma: Real
            Euler-Maruyama discretization step of the Langevin equation (see `Notes`).
        """
        self._f = f
        self._beta = f.diff_lipschitz
        self._gamma = self._set_gamma(gamma)
        self._rng = None
        self.x = None

    def _sample_init(self, rng, x0: pxt.NDArray):
        r"""
        Parameters
        ----------
        rng:
            Internal random generator.
        x0: NDArray
            Starting point of the Markov chain.
        """
        self.x = x0.copy()
        if rng is None:
            xp = pxu.get_array_module(x0)
            self._rng = xp.random.default_rng(None)
        else:
            self._rng = rng

    @pxrt.enforce_precision()
    def _sample(self) -> pxt.NDArray:
        x = self.x.copy()
        x -= self._gamma * pxu.copy_if_unsafe(self._f.grad(self.x))
        x += math.sqrt(2 * self._gamma) * self._rng.standard_normal(size=self.x.shape, dtype=self.x.dtype)
        self.x = x
        return x

    def objective_func(self) -> pxt.Real:
        r"""
        Negative logarithm of the target ditribution (up to the a constant) evaluated at the current state of the Markov
        chain.

        Useful for diagnostics purposes to monitor whether the Markov chain is sufficiently warm-started.  If so, the
        samples should accumulate around the modes of the target distribution, i.e., toward the minimum of
        :math:`\mathcal{F}`.
        """
        return pxu.copy_if_unsafe(self._f.apply(self.x))

    def _set_gamma(self, gamma: pxt.Real = None) -> pxt.Real:
        if gamma is None:
            if math.isfinite(self._beta):
                return pxrt.coerce(0.98 / self._beta)
            else:
                msg = "If f has unbounded Lipschitz gradient, the gamma parameter must be provided."
            raise ValueError(msg)
        else:
            try:
                assert gamma > 0
            except Exception:
                raise ValueError(f"gamma must be positive, got {gamma}.")
            return pxrt.coerce(gamma)


class MYULA(ULA):
    r"""
    Moreau-Yosida unajusted Langevin algorithm (MYULA).

    Generates samples from the distribution

    .. math::

       p(\mathbf{x}) = \frac{\exp(-\mathcal{F}(\mathbf{x}) - \mathcal{G}(\mathbf{x}))}{\int_{\mathbb{R}^N}
       \exp(-\mathcal{F}(\tilde{\mathbf{x}}) - \mathcal{G}(\tilde{\mathbf{x}})) \mathrm{d} \tilde{\mathbf{x}} },

    where :math:`\mathcal{F}: \mathbb{R}^N \to \mathbb{R}` is *convex* and *differentiable* with :math:`\beta`-
    *Lipschitz continuous* gradient, and :math:`\mathcal{G}: \mathbb{R}^N \to \mathbb{R}` is *proper*, *lower semi-
    continuous* and *convex* with *simple proximal operator*.

    Notes
    -----
    MYULA is an extension of :py:class:`~pyxu.experimental.sampler.ULA` to sample from distributions whose logarithm is
    nonsmooth.  It consists in applying ULA to the differentiable functional :math:`\mathcal{U}^\lambda = \mathcal{F} +
    \mathcal{G}^\lambda` for some :math:`\lambda > 0`, where

     .. math::

        \mathcal{G}^\lambda (\mathbf{x}) = \inf_{\tilde{\mathbf{x}} \in \mathbb{R}^N} \frac{1}{2 \lambda} \Vert
        \tilde{\mathbf{x}} - \mathbf{x} \Vert_2^2 + \mathcal{G}(\tilde{\mathbf{x}})

    is the Moreau-Yosida envelope of :math:`\mathcal{G}` with parameter :math:`\lambda`.  We then have

    .. math::

       \nabla \mathcal{U}^\lambda (\mathbf{x}) = \nabla \mathcal{F}(\mathbf{x}) + \frac{1}{\lambda} (\mathbf{x} -
       \mathrm{prox}_{\lambda \mathcal{G}}(\mathbf{x})),

    hence :math:`\nabla \mathcal{U}^\lambda` is :math:`(\beta + \frac{1}{\lambda})`-Lipschitz continuous, where
    :math:`\beta` is the Lipschitz constant of :math:`\nabla \mathcal{F}`.  Note that the target distribution of the
    underlying ULA Markov chain is not exactly :math:`p(\mathbf{x})`, but the distribution

    .. math::

       p^\lambda(\mathbf{x}) \propto \exp(-\mathcal{F}(\mathbf{x})-\mathcal{G}^\lambda(\mathbf{x})),

    which introduces some additional bias on top of the bias of ULA related to the step size :math:`\gamma` (see `Notes`
    of :py:class:`~pyxu.experimental.sampler.ULA` documentation).  MYULA is guaranteed to converges when :math:`\gamma
    \leq \frac{1}{\beta + \frac{1}{\lambda}}`, in which case it converges toward the stationary distribution
    :math:`p^\lambda_\gamma(\mathbf{x})` that satisfies

    .. math::

       \lim_{\gamma, \lambda \to 0} \Vert p^\lambda_\gamma - p \Vert_{\mathrm{TV}} = 0

    (see [MYULA]_).  The parameter :math:`\lambda` parameter is subject to a similar bias-variance tradeoff as
    :math:`\gamma`.  It is recommended to set it in the order of :math:`\frac{1}{\beta}`, so that the contributions of
    :math:`\mathcal{F}` and :math:`\mathcal{G}^\lambda` to the Lipschitz constant of :math:`\nabla \mathcal{U}^\lambda`
    is well balanced.
    """

    def __init__(
        self,
        f: pxa.DiffFunc = None,
        g: pxa.ProxFunc = None,
        gamma: pxt.Real = None,
        lamb: pxt.Real = None,
    ):
        r"""
        Parameters
        ----------
        f: :py:class:`~pyxu.abc.DiffFunc`, None
            Differentiable functional.
        g: :py:class:`~pyxu.abc.ProxFunc`, None
            Proximable functional.
        gamma: Real
            Euler-Maruyama discretization step of the Langevin equation (see `Notes` of
            :py:class:`~pyxu.experimental.sampler.ULA` documentation).
        lamb: Real
            Moreau-Yosida envelope parameter for `g`.
        """
        dim = None
        if f is not None:
            dim = f.dim
        if g is not None:
            if dim is None:
                dim = g.dim
            else:
                assert g.dim == dim
        if dim is None:
            raise ValueError("One of f or g must be nonzero.")

        self._f_diff = pxo.NullFunc(dim=dim) if (f is None) else f
        self._g = pxo.NullFunc(dim=dim) if (g is None) else g

        self._lambda = self._set_lambda(lamb)
        f = self._f_diff + self._g.moreau_envelope(self._lambda)
        f.diff_lipschitz = f.estimate_diff_lipschitz()
        super().__init__(f, gamma)

    def _set_lambda(self, lamb: pxt.Real = None) -> pxt.Real:
        if lamb is None:
            if self._g._name == "NullFunc":
                return pxrt.coerce(1)  # Lambda is irrelevant if g is a NullFunc, but it must be positive
            elif math.isfinite(dl := self._f_diff.diff_lipschitz):
                return pxrt.coerce(2) if dl == 0 else pxrt.coerce(min(2, 1 / dl))
            else:
                msg = "If f has unbounded Lipschitz gradient, the lambda parameter must be provided."
            raise ValueError(msg)
        else:
            return pxrt.coerce(lamb)



import itertools

import numpy as np
import pytest
from test_diff import DiffOpMixin, apply_gradient, apply_hessian

import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu_tests.operator.conftest as conftest


class DirDerOpMixin(DiffOpMixin):
    @pytest.fixture(
        params=[
            #          (arg_shape, directions)
            (
                (5,),
                ((1.0,),),
            ),
            (
                (5, 5, 5),
                ((0.1, 2.0, 1.0),),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        return _spec[0]

    @pytest.fixture
    def diff_method(self):
        return "fd"

    @pytest.fixture(params=["constant", "varying"])
    def directions(self, _spec, arg_shape, request):
        dirs = np.array(_spec[1])
        for i in range(len(dirs)):
            dirs[i] = dirs[i] / np.linalg.norm(dirs[i])

        if request.param == "constant":
            pass
        else:
            expand_dims = -np.arange(len(arg_shape)) - 1
            broadcast_arr = np.ones(
                (
                    1,
                    len(arg_shape),
                )
                + arg_shape
            )
            dirs = np.expand_dims(dirs, expand_dims.tolist()) * broadcast_arr

        return dirs


class TestDirectionalDerivative(DirDerOpMixin):
    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        sh = (size, size)
        return sh

    @pytest.fixture(params=[1, 2])
    def order(self, request):
        return request.param

    @pytest.fixture
    def diff_kwargs(self, arg_shape, order, directions, ndi, width, sampling, diff_method):
        xp = ndi.module()
        return {
            "arg_shape": arg_shape,
            "order": order,
            "directions": xp.array(directions[0], dtype=width.value),
            "mode": "constant",
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def diff_op(self):
        return pxo.DirectionalDerivative

    @pytest.fixture
    def data_apply(self, op, arg_shape, diff_method, gt_diffs, directions, order) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=20)
        out = np.zeros(arg_shape)
        grad_directions = np.arange(len(arg_shape), dtype=int)
        if order == 1:
            grad = apply_gradient(arr, arg_shape, gt_diffs, grad_directions, diff_method, mode="constant")
            for i in range(len(arg_shape)):
                out += directions[0][i] * grad[i].squeeze()
        else:  # order == 2
            hess_directions = tuple(list(_) for _ in itertools.combinations_with_replacement(grad_directions, 2))
            hess = apply_hessian(arr, arg_shape, gt_diffs, hess_directions, diff_method, mode="constant")
            c = 0
            for i in range(len(arg_shape)):
                for j in range(i, len(arg_shape)):
                    factor = 1 if i == j else 2
                    out += directions[0][i] * hess[c].squeeze() * directions[0][j] * factor
                    c += 1

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestDirectionalGradient(DirDerOpMixin):
    @pytest.fixture(
        params=[
            #          (arg_shape, directions)
            (
                (5, 5, 5),
                (
                    (0.1, 2.0, 1.0),
                    (0.1, 1.0, 2.0),
                    (2.0, 0.1, 1.0),
                    (2.0, 1.0, 0.1),
                ),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def data_shape(self, arg_shape, directions) -> pxt.NDArrayShape:
        size = len(directions) * np.prod(arg_shape).item()
        sh = (size, size)
        return sh

    @pytest.fixture
    def diff_kwargs(self, arg_shape, directions, ndi, width, sampling, diff_method):
        xp = ndi.module()
        return {
            "arg_shape": arg_shape,
            "directions": [xp.array(direction, dtype=width.value) for direction in directions],
            "mode": "constant",
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def diff_op(self):
        return pxo.DirectionalGradient

    @pytest.fixture
    def data_apply(self, op, arg_shape, diff_method, gt_diffs, directions) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=20)
        out = np.zeros((len(directions), *arg_shape))
        grad_directions = np.arange(len(arg_shape), dtype=int)
        grad = apply_gradient(arr, arg_shape, gt_diffs, grad_directions, diff_method, mode="constant")
        for j, direction in enumerate(directions):
            for i in range(len(arg_shape)):
                out[j] += direction[i] * grad[i].squeeze()

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestDirectionalLaplacian(DirDerOpMixin):
    @pytest.fixture(
        params=[
            #          (arg_shape, directions)
            (
                (5, 5, 5),
                (
                    (0.1, 2.0, 1.0),
                    (0.1, 1.0, 2.0),
                    (2.0, 0.1, 1.0),
                    (2.0, 1.0, 0.1),
                ),
                (0.1, 0.2, 0.3, 0.5),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def weights(self, _spec):
        return _spec[2]

    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        sh = (size, size)
        return sh

    @pytest.fixture
    def diff_kwargs(self, arg_shape, directions, ndi, width, sampling, diff_method, weights):
        xp = ndi.module()
        return {
            "arg_shape": arg_shape,
            "directions": [xp.array(direction, dtype=width.value) for direction in directions],
            "mode": "constant",
            "sampling": sampling,
            "diff_method": diff_method,
            "weights": weights,
        }

    @pytest.fixture
    def diff_op(self):
        return pxo.DirectionalLaplacian

    @pytest.fixture
    def data_apply(self, op, arg_shape, diff_method, gt_diffs, directions, weights) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=20)
        out = np.zeros(arg_shape)
        grad_directions = np.arange(len(arg_shape), dtype=int)
        hess_directions = tuple(list(_) for _ in itertools.combinations_with_replacement(grad_directions, 2))
        hess = apply_hessian(arr, arg_shape, gt_diffs, hess_directions, diff_method, mode="constant")
        for weight, direction in zip(weights, directions):
            c = 0
            for i in range(len(arg_shape)):
                for j in range(i, len(arg_shape)):
                    factor = 1 if i == j else 2
                    out += weight * direction[i] * hess[c].squeeze() * direction[j] * factor
                    c += 1

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestDirectionalHessian(DirDerOpMixin):
    @pytest.fixture(
        params=[
            #          (arg_shape, directions)
            (
                (5, 5, 5),
                (
                    (0.1, 2.0, 1.0),
                    (0.1, 1.0, 2.0),
                    (2.0, 0.1, 1.0),
                    (2.0, 1.0, 0.1),
                ),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def data_shape(self, arg_shape, directions) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        ndim_hess = len(directions) * (len(directions) + 1) // 2
        sh = (ndim_hess * size, size)
        return sh

    @pytest.fixture
    def diff_kwargs(self, arg_shape, directions, ndi, width, sampling, diff_method):
        xp = ndi.module()
        return {
            "arg_shape": arg_shape,
            "directions": [xp.array(direction, dtype=width.value) for direction in directions],
            "mode": "constant",
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def diff_op(self):
        return pxo.DirectionalHessian

    @pytest.fixture
    def data_apply(self, op, arg_shape, diff_method, gt_diffs, directions) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=20)
        grad_directions = np.arange(len(arg_shape), dtype=int)
        hess_directions = tuple(list(_) for _ in itertools.combinations_with_replacement(grad_directions, 2))
        dir_hess_directions = tuple(
            list(_) for _ in itertools.combinations_with_replacement(np.arange(len(directions)), 2)
        )
        out = np.zeros((len(dir_hess_directions), *arg_shape))
        hess = apply_hessian(arr, arg_shape, gt_diffs, hess_directions, diff_method, mode="constant")
        c1 = 0
        for k1, direction1 in enumerate(directions):
            for k2, direction2 in enumerate(directions[k1:]):
                c2 = 0
                for i in range(len(arg_shape)):
                    for j in range(i, len(arg_shape)):
                        factor = 1 if i == j else 2
                        out[c1] += direction1[i] * hess[c2].squeeze() * direction2[j] * factor
                        c2 += 1
                c1 += 1
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )
